################################################################################
#                                                                              #
#  Copyright 2015-2019 Max Planck Institute for Dynamics and Self-Organization #
#                                                                              #
#  This file is part of TurTLE.                                                #
#                                                                              #
#  TurTLE is free software: you can redistribute it and/or modify              #
#  it under the terms of the GNU General Public License as published           #
#  by the Free Software Foundation, either version 3 of the License,           #
#  or (at your option) any later version.                                      #
#                                                                              #
#  TurTLE is distributed in the hope that it will be useful,                   #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with TurTLE.  If not, see <http://www.gnu.org/licenses/>              #
#                                                                              #
# Contact: Cristian.Lalescu@ds.mpg.de                                          #
#                                                                              #
################################################################################



import os
import h5py
import numpy as np

import TurTLE
from ._code import _code
from TurTLE import tools

class DNS(_code):
    """This class is meant to stitch together the C++ code into a final source file,
    compile it, and handle all job launching.
    """
    from .DNS_statistics import \
        compute_Reynolds_stress_invariants, \
        compute_statistics, \
        compute_statistics_error, \
        compute_time_averages, \
        set_plt_style, \
        read_moments, \
        read_histogram, \
        plot_basic_stats
    def __init__(
            self,
            work_dir = './',
            simname = 'test'):
        _code.__init__(
                self,
                work_dir = work_dir,
                simname = simname)
        self.list_of_particle_codes = [
                'NSVEparticles_no_output',
                'NSVEcomplex_particles',
                'NSVE_Stokes_particles',
                'NSVEparticles',
                'NSVEp_extra_sampling',
                'static_field',
                'static_field_with_ghost_collisions',
                'kraichnan_field']
        self.extra_parameters = {}
        self.statistics = {}
        return None
    def write_src(
            self):
        self.version_message = (
                '/***********************************************************************\n' +
                '* this code automatically generated by TurTLE\n' +
                '* version {0}\n'.format(TurTLE.__version__) +
                '***********************************************************************/\n\n\n')
        self.include_list = [
                '"full_code/main_code.hpp"',
                '"full_code/{0}.hpp"\n'.format(self.dns_type)]
        self.main = """
            int main(int argc, char *argv[])
            {{
                bool fpe = (
                    (getenv("TURTLE_FPE_OFF") == nullptr) ||
                    (getenv("TURTLE_FPE_OFF") != std::string("TRUE")));
                return main_code< {0} >(argc, argv, fpe);
            }}
            """.format(self.dns_type + '<{0}>'.format(self.C_field_dtype))
        self.includes = '\n'.join(
                ['#include ' + hh
                 for hh in self.include_list])
        with open(self.name + '.cpp', 'w') as outfile:
            outfile.write(self.version_message + '\n\n')
            outfile.write(self.includes + '\n\n')
            outfile.write(self.main + '\n')
        return None
    def generate_default_parameters(self):
        # these parameters are relevant for all DNS classes
        self.parameters['fftw_plan_rigor'] = str('FFTW_ESTIMATE')
        self.parameter_description['fftw_plan_rigor'] = 'FFTW plan rigor to use. One of `FFTW_ESTIMATE`, `FFTW_MEASURE`, `FFTW_PATIENT`. Please see FFTW documentation.'
        self.parameters['dealias_type'] = int(1)
        self.parameter_description['dealias_type'] = 'Dealiasing mehtod to use, integer. Options are: two-thirds (0) or smooth (1).'
        self.parameters['dkx'] = float(1.0)
        self.parameter_description['dkx'] = 'Smallest wavenumber in the x direction for a pseudo-spectral run.'
        self.parameters['dky'] = float(1.0)
        self.parameter_description['dky'] = 'Smallest wavenumber in the y direction for a pseudo-spectral run.'
        self.parameters['dkz'] = float(1.0)
        self.parameter_description['dkz'] = 'Smallest wavenumber in the z direction for a pseudo-spectral run.'
        self.parameters['dt'] = float(0.01)
        self.parameter_description['dt'] = 'Fixed timestep to use. It is strongly recommended not to change this value in between jobs.'
        self.parameters['nu'] = float(0.1)
        self.parameter_description['nu'] = 'Viscosity value used in the equations, given in code units.'
        self.parameters['fmode'] = int(1)
        self.parameter_description['fmode'] = 'Forcing parameter: mode to use for the Kolmogorov forcing.'
        self.parameters['famplitude'] = float(0.5)
        self.parameter_description['famplitude'] = 'Forcing parameter: amplitude of Kolmogorov forcing, in code units.'
        self.parameters['friction_coefficient'] = float(0.5)
        self.parameter_description['friction_coefficient'] = 'Forcing parameter: drag coefficient, in code units.'
        self.parameters['variation_strength'] = float(0.25)
        self.parameter_description['variation_strength'] = 'Forcing parameter: amplitude of time variation of injection rate, in code units.'
        self.parameters['variation_time_scale'] = float(1.)
        self.parameter_description['variation_time_scale'] = 'Forcing parameter: time scale of variation of injection rate, in code units.'
        self.parameters['energy'] = float(0.5)
        self.parameter_description['energy'] = 'Forcing parameter: if fluid is forced by enforcing a constant energy, this is the value (in code units).'
        self.parameters['injection_rate'] = float(0.4)
        self.parameter_description['injection_rate'] = 'Forcing parameter: if a fixed energy injection rate is used, this is the value (in code units).'
        self.parameters['fk0'] = float(2.0)
        self.parameter_description['fk0'] = 'Forcing parameter: if forcing acts on wavenumber band, this is the smallest wavenumber where it acts (in code units).'
        self.parameters['fk1'] = float(4.0)
        self.parameter_description['fk1'] = 'Forcing parameter: if forcing acts on wavenumber band, this is the largest wavenumber where it acts (in code units).'
        self.parameters['forcing_type'] = str('fixed_energy_injection_rate')
        self.parameter_description['forcing_type'] = 'Forcing parameter: what type of force to use.'
        self.parameters['histogram_bins'] = int(256)
        self.parameter_description['histogram_bins'] = 'During statistics, histograms of real-valued fields are computed using a number of `HISTOGRAM_BINS` bins.'
        self.parameters['max_velocity_estimate'] = float(1)
        self.parameter_description['max_velocity_estimate'] = 'During statistics, velocity histogram bins are computed using this estimate (see code for details).'
        self.parameters['max_vorticity_estimate'] = float(1)
        self.parameter_description['max_velocity_estimate'] = 'During statistics, vorticity histogram bins are computed using this estimate (see code for details).'
        self.parameters['field_random_seed'] = int(0)
        self.parameter_description['field_random_seed'] = 'Seed for random number generator used for initial field. 0 means "python-generated field".'
        # parameters specific to particle version
        self.NSVEp_extra_parameters = {}
        self.NSVEp_extra_parameters['niter_part'] = int(1)
        self.NSVEp_extra_parameters['niter_part_fine_period'] = int(10)
        self.NSVEp_extra_parameters['niter_part_fine_duration'] = int(0)
        self.NSVEp_extra_parameters['nparticles'] = int(10)
        self.NSVEp_extra_parameters['cpp_random_particles'] = int(1)
        self.NSVEp_extra_parameters['sample_acceleration'] = int(0)
        self.NSVEp_extra_parameters['tracers0_integration_steps'] = int(4)
        self.NSVEp_extra_parameters['tracers0_neighbours'] = int(1)
        self.NSVEp_extra_parameters['tracers0_smoothness'] = int(1)
        self.NSVEp_extra_parameters['tracers0_enable_p2p'] = int(0)
        self.NSVEp_extra_parameters['tracers0_enable_inner'] = int(0)
        self.NSVEp_extra_parameters['tracers0_enable_vorticity_omega'] = int(0)
        self.NSVEp_extra_parameters['tracers0_cutoff'] = float(1)
        self.NSVEp_extra_parameters['tracers0_inner_v0'] = float(1)
        self.NSVEp_extra_parameters['tracers0_lambda'] = float(1)
        return None

    def generate_extra_parameters(self):
        # this is to ensure that particle codes have the dictionaries defined
        for kk in self.list_of_particle_codes:
            if kk not in self.extra_parameters.keys():
                self.extra_parameters[kk] = {}
        self.extra_parameters['kraichnan_field'].update({
            'output_velocity': int(1),
            'spectrum_dissipation': float(0.027),
            'spectrum_Lint': float(1.3),
            'spectrum_etaK': float(0.3),
            'spectrum_large_scale_const': float(6.78),
            'spectrum_small_scale_const': float(0.40)})
        self.extra_parameters['NSVE_Stokes_particles'].update({
            'initial_field_amplitude': float(0.0),
            'initial_particle_vel': float(0.05),
            'drag_coefficient': float(0.1)})
        self.extra_parameters['NSVEp_extra_sampling'].update({
            'sample_pressure': int(1),
            'sample_pressure_gradient': int(1),
            'sample_pressure_Hessian': int(1),
            'sample_velocity_gradient': int(1)})
        return None

    def get_kspace(self):
        kspace = {}
        if self.parameters['dealias_type'] == 1:
            kMx = self.parameters['dkx']*(self.parameters['nx']//2 - 1)
            kMy = self.parameters['dky']*(self.parameters['ny']//2 - 1)
            kMz = self.parameters['dkz']*(self.parameters['nz']//2 - 1)
        else:
            kMx = self.parameters['dkx']*(self.parameters['nx']//3 - 1)
            kMy = self.parameters['dky']*(self.parameters['ny']//3 - 1)
            kMz = self.parameters['dkz']*(self.parameters['nz']//3 - 1)
        kspace['kM'] = max(kMx, kMy, kMz)
        kspace['dk'] = min(self.parameters['dkx'],
                           self.parameters['dky'],
                           self.parameters['dkz'])
        nshells = int(kspace['kM'] / kspace['dk']) + 2
        kspace['nshell'] = np.zeros(nshells, dtype = np.int64)
        kspace['kshell'] = np.zeros(nshells, dtype = np.float64)
        kspace['kx'] = np.arange( 0,
                                  self.parameters['nx']//2 + 1).astype(np.float64)*self.parameters['dkx']
        kspace['ky'] = np.arange(-self.parameters['ny']//2 + 1,
                                  self.parameters['ny']//2 + 1).astype(np.float64)*self.parameters['dky']
        kspace['ky'] = np.roll(kspace['ky'], self.parameters['ny']//2+1)
        kspace['kz'] = np.arange(-self.parameters['nz']//2 + 1,
                                  self.parameters['nz']//2 + 1).astype(np.float64)*self.parameters['dkz']
        kspace['kz'] = np.roll(kspace['kz'], self.parameters['nz']//2+1)
        return kspace
    def convert_complex_from_binary(
            self,
            field_name = 'vorticity',
            iteration = 0,
            file_name = None):
        """read the Fourier representation of a vector field.

        Read the binary file containing iteration ``iteration`` of the
        field ``field_name``, and write it in a ``.h5`` file.
        """
        data = np.memmap(
                os.path.join(self.work_dir,
                             self.simname + '_{0}_i{1:0>5x}'.format('c' + field_name, iteration)),
                dtype = self.ctype,
                mode = 'r',
                shape = (self.parameters['ny'],
                         self.parameters['nz'],
                         self.parameters['nx']//2+1,
                         3))
        if type(file_name) == type(None):
            file_name = self.simname + '_{0}_i{1:0>5x}.h5'.format('c' + field_name, iteration)
            file_name = os.path.join(self.work_dir, file_name)
        f = h5py.File(file_name, 'a')
        f[field_name + '/complex/{0}'.format(iteration)] = data
        f.close()
        return None
    def write_par(
            self,
            iter0 = 0):
        assert (self.parameters['niter_todo'] % self.parameters['niter_stat'] == 0)
        assert (self.parameters['niter_todo'] % self.parameters['niter_out']  == 0)
        assert (self.parameters['niter_out']  % self.parameters['niter_stat'] == 0)
        if self.dns_type in self.list_of_particle_codes:
            assert (self.parameters['niter_todo'] % self.parameters['niter_part'] == 0)
            assert (self.parameters['niter_out']  % self.parameters['niter_part'] == 0)
        _code.write_par(self, iter0 = iter0)
        with h5py.File(self.get_data_file_name(), 'r+') as ofile:
            ofile['code_info/exec_name'] = self.name
            kspace = self.get_kspace()
            for k in kspace.keys():
                ofile['kspace/' + k] = kspace[k]
            nshells = kspace['nshell'].shape[0]
            kspace = self.get_kspace()
            nshells = kspace['nshell'].shape[0]
            vec_stat_datasets = ['velocity', 'vorticity']
            scal_stat_datasets = []
            time_chunk = 2**17 # 2^17 8-byte values, i.e. 1 megabyte
            ofile.create_dataset('statistics/CFL_velocity',
                                 (1,),
                                 chunks = (time_chunk,),
                                 maxshape = (None,),
                                 dtype = np.float64)
            for k in vec_stat_datasets:
                time_chunk = 2**20//(8*3*3*nshells)
                time_chunk = max(time_chunk, 1)
                ofile.create_dataset('statistics/spectra/' + k + '_' + k,
                                     (1, nshells, 3, 3),
                                     chunks = (time_chunk, nshells, 3, 3),
                                     maxshape = (None, nshells, 3, 3),
                                     dtype = np.float64)
                time_chunk = 2**20//(8*4*10)
                time_chunk = max(time_chunk, 1)
                a = ofile.create_dataset('statistics/moments/' + k,
                                     (1, 10, 4),
                                     chunks = (time_chunk, 10, 4),
                                     maxshape = (None, 10, 4),
                                     dtype = np.float64)
                time_chunk = 2**20//(8*4*self.parameters['histogram_bins'])
                time_chunk = max(time_chunk, 1)
                ofile.create_dataset('statistics/histograms/' + k,
                                     (1,
                                      self.parameters['histogram_bins'],
                                      4),
                                     chunks = (time_chunk,
                                               self.parameters['histogram_bins'],
                                               4),
                                     maxshape = (None,
                                                 self.parameters['histogram_bins'],
                                                 4),
                                     dtype = np.int64)
            ofile['checkpoint'] = int(0)
            if self.dns_type in ['static_field_with_ghost_collisions']:
                ofile.create_group('statistics/collisions')
        if (self.dns_type in ['NSE', 'NSE_alt_dealias', 'NSVE', 'NSVE_no_output']):
            return None
        return None
    def simulation_parser_arguments(
            self,
            parser):
        parser.add_argument(
                '--simname',
                type = str, dest = 'simname',
                default = 'test')
        parser.add_argument(
               '-n', '--grid-size',
               type = int,
               dest = 'n',
               default = 32,
               metavar = 'N',
               help = 'code is run by default in a grid of NxNxN')
        for coord in ['x', 'y', 'z']:
            parser.add_argument(
                   '--L{0}'.format(coord), '--box-length-{0}'.format(coord),
                   type = float,
                   dest = 'L{0}'.format(coord),
                   default = 2.0,
                   metavar = 'length{0}'.format(coord),
                   help = 'length of the box in the {0} direction will be `length{0} x pi`'.format(coord))
        parser.add_argument(
                '--wd',
                type = str, dest = 'work_dir',
                default = './')
        parser.add_argument(
                '--precision',
                choices = ['single', 'double'],
                type = str,
                default = 'single')
        parser.add_argument(
                '--src-wd',
                type = str,
                dest = 'src_work_dir',
                help = ('Empty by default.'
                    + ' Location of simulation from which to source initial conditions.'
                    + ' See also `--src-simname`.'),
                default = '')
        parser.add_argument(
                '--src-simname',
                type = str,
                dest = 'src_simname',
                help = ('Empty by default.'
                    + ' Name of simulation from which to source initial conditions.'
                    + ' If the parameter file of the source simulation is present, '
                    + 'the new DNS will use the same physical parameters '
                    + '(forcing parameters and viscosity) and the same dealiasing option.'
                    + ' Any other parameters (e.g. "nx") have to be specified separately.'
                    + ' See also `--overwrite-src-parameters`.'),
                default = '')
        parser.add_argument(
                '--src-iteration',
                type = int,
                dest = 'src_iteration',
                help = ('0 by default.'
                    + ' Iteration of source simulation to be used as initial condition.'
                    + ' See also `--src-simname`.'),
                default = 0)
        parser.add_argument(
                '--overwrite-src-parameters',
                action = 'store_true',
                dest = 'overwrite_src_parameters',
                help = ('False by default.'
                      + ' If using a source simulation, do we keep its physical parameters?'
                      + ' Note: if this option is switched on, default values are used for '
                      + 'parameters that are not set explicitly by the command --- '
                      + 'NOT the values of the source simulation.'))
        parser.add_argument(
               '--kMeta',
               type = float,
               dest = 'kMeta',
               default = 2.0)
        parser.add_argument(
               '--dtfactor',
               type = float,
               dest = 'dtfactor',
               default = 0.5,
               help = 'dt is computed as DTFACTOR / N')
        return None
    def particle_parser_arguments(
            self,
            parser):
        parser.add_argument(
               '--particle-rand-seed',
               type = int,
               dest = 'particle_rand_seed',
               default = None)
        parser.add_argument(
               '--pclouds',
               type = int,
               dest = 'pclouds',
               default = 1,
               help = ('number of particle clouds. Particle "clouds" '
                       'consist of particles distributed according to '
                       'pcloud-type.'))
        parser.add_argument(
                '--pcloud-type',
                choices = ['random-cube',
                           'regular-cube'],
                dest = 'pcloud_type',
                default = 'random-cube')
        parser.add_argument(
               '--particle-cloud-size',
               type = float,
               dest = 'particle_cloud_size',
               default = 2*np.pi)
        return None
    def add_parser_arguments(
            self,
            parser):
        """Creates subparsers for all simulation types and supplies them with parameters"""
        subparsers = parser.add_subparsers(
                dest = 'DNS_class',
                help = 'type of simulation to run')
        subparsers.required = True
        parser_list = {}

        parser_list['NSE'] = subparsers.add_parser(
                'NSE',
                help = 'plain Navier-Stokes equation')

        parser_list['NSE_alt_dealias'] = subparsers.add_parser(
                'NSE_alt_dealias',
                help = 'plain Navier-Stokes equation alternate dealiasing')

        parser_list['NSVE'] = subparsers.add_parser(
                'NSVE',
                help = 'plain Navier-Stokes vorticity equation')

        parser_list['NSVE_no_output'] = subparsers.add_parser(
                'NSVE_no_output',
                help = 'plain Navier-Stokes vorticity formulation, checkpoints are NOT SAVED')

        parser_list['NSVEparticles_no_output'] = subparsers.add_parser(
                'NSVEparticles_no_output',
                help = 'plain Navier-Stokes vorticity formulation, with basic fluid tracers, checkpoints are NOT SAVED')

        parser_list['static_field'] = subparsers.add_parser(
                'static_field',
                help = 'static field with basic fluid tracers')

        parser_list['static_field_with_ghost_collisions'] = subparsers.add_parser(
                'static_field_with_ghost_collisions',
                help = 'static field with basic fluid tracers and ghost collisions')

        parser_list['kraichnan_field'] = subparsers.add_parser(
                'kraichnan_field',
                help = 'Kraichnan field with basic fluid tracers')

        parser_list['NSVEparticles'] = subparsers.add_parser(
                'NSVEparticles',
                help = 'plain Navier-Stokes vorticity formulation, with basic fluid tracers')

        parser_list['NSVE_Stokes_particles'] = subparsers.add_parser(
                'NSVE_Stokes_particles',
                help = 'plain Navier-Stokes vorticity formulation, with passive Stokes drag particles')

        parser_list['NSVEcomplex_particles'] = subparsers.add_parser(
                'NSVEcomplex_particles',
                help = 'plain Navier-Stokes vorticity formulation, with oriented active particles')
        parser_list['NSVEp_extra_sampling'] = subparsers.add_parser(
                'NSVEp_extra_sampling',
                help = 'plain Navier-Stokes vorticity formulation, with basic fluid tracers, that sample velocity gradient, as well as pressure and its derivatives.')

        for extra_dns_type in self.extra_parameters.keys():
            if extra_dns_type not in parser_list.keys():
                parser_list[extra_dns_type] = subparsers.add_parser(
                        extra_dns_type,
                        help = 'Custom DNS class of type ' + extra_dns_type)

        for pp in parser_list:
            self.simulation_parser_arguments(parser_list[pp])
            self.job_parser_arguments(parser_list[pp])
            self.parameters_to_parser_arguments(parser_list[pp])
            if pp in self.extra_parameters:
                self.parameters_to_parser_arguments(
                        parser_list[pp],
                        self.extra_parameters[pp])

        for pp in self.list_of_particle_codes:
            self.particle_parser_arguments(parser_list[pp])
            #self.parameters_to_parser_arguments(parser_list[pp],
            #        self.NSVEp_extra_parameters)
        return None

    def custom_parameter_update(self):
        """To be reimplemented by child classes.

            Children of `DNS` should modify the default values of
            parameters in reimplementations of this method.
            The method is called *before* parameters are read from
            the command line.
        """
        return None

    def prepare_launch(
            self,
            args = [],
            extra_parameters = None):
        """Assign DNS parameters.

        There are three ways to specify parameters:

        * Directly via the command line.
        * If the simulation takes its initial condition from another simulation
        (via the :code:`--src-dir` and :code:`--src-simname` parameters) and
        some physical parameter is not provided on the command line, then this
        method will attempt to read that parameter from the parameter file of
        the source simulation.
        * If a parameter is not specified via the command line, and a source
        simulation is not provided, or the source parameter file is missing, or
        the :code:`--overwrite-src-parameters` option is turned on, this method
        will generate a "reasonable" value for that parameter.
        "Reasonable" refers to either some fixed values for the forcing
        parameters, or an empirical extrapolation of the viscosity to satisfy a
        given value of the :code:`--kMeta` command line parameter.

        With the default Lundgren forcing applied in the band [2, 4],
        we can estimate the dissipation, therefore we can estimate
        :math:`k_M \\eta_K` and constrain the viscosity.

        In brief, the command line parameter :math:`k_M \\eta_K` is
        used in the following formula for :math:`\\nu` (:math:`N` is the
        number of real space grid points per coordinate):

        .. math::

            \\nu = \\left(\\frac{2 k_M \\eta_K}{N} \\right)^{4/3}

        With this choice, the average dissipation :math:`\\varepsilon`
        will be close to 0.4, and the integral scale velocity will be
        close to 0.77, yielding the approximate value for the Taylor
        microscale and corresponding Reynolds number:

        .. math::

            \\lambda \\approx 4.75\\left(\\frac{2 k_M \\eta_K}{N} \\right)^{4/6}, \\hskip .5in
            R_\\lambda \\approx 3.7 \\left(\\frac{N}{2 k_M \\eta_K} \\right)^{4/6}

        """
        self.generate_default_parameters()
        self.generate_extra_parameters()
        for key in self.list_of_particle_codes:
            assert(key in self.extra_parameters.keys())
            self.extra_parameters[key].update(self.NSVEp_extra_parameters)
        self.custom_parameter_update()

        opt = _code.prepare_launch(self, args = args)
        self.set_precision(opt.precision)
        self.dns_type = opt.DNS_class
        self.name = self.dns_type + '-' + self.fluid_precision + '-v' + TurTLE.__version__

        # now add parameters for the chosen dns_type to self.parameters
        # extra parameters from prepare launch argument
        if type(extra_parameters) != type(None):
            if self.dns_type in extra_parameters.keys():
                for k in extra_parameters[self.dns_type].keys():
                    self.parameters[k] = extra_parameters[self.dns_type][k]
        # extra parameters from `self.extra_parameters`
        if self.dns_type in self.extra_parameters.keys():
            for k in self.extra_parameters[self.dns_type].keys():
                self.parameters[k] = self.extra_parameters[self.dns_type][k]

        # for parameters there are three options, in order of preference:
        # * value given from command line
        # * value used by source simulation
        # * default value

        if ((self.parameters['niter_todo'] % self.parameters['niter_out']) != 0):
            self.parameters['niter_out'] = self.parameters['niter_todo']

        if len(opt.src_simname) > 0:
            if len(opt.src_work_dir) == 0:
                opt.src_work_dir = os.path.realpath(opt.work_dir)
            src_file_name = os.path.join(
                    os.path.realpath(opt.src_work_dir),
                    opt.src_simname + '.h5')
            if os.path.exists(src_file_name):
                # open source file for use later
                src_file = h5py.File(src_file_name, 'r')
            else:
                print('TurTLE WARNING: unable to copy parameters of source simulation.'
                    + 'File {0} is missing.\n'.format(src_file_name)
                    + 'DNS will use default values for physical parameters.\n')
                src_file = None
        else:
            src_file = None
        if opt.overwrite_src_parameters:
            src_file = None
        if type(opt.dkx) == type(None):
            opt.dkx = 2. / opt.Lx
        if type(opt.dky) == type(None):
            opt.dky = 2. / opt.Ly
        if type(opt.dkz) == type(None):
            opt.dkz = 2. / opt.Lz
        if type(opt.nx) == type(None):
            opt.nx = opt.n
        if type(opt.ny) == type(None):
            opt.ny = opt.n
        if type(opt.nz) == type(None):
            opt.nz = opt.n
        if type(opt.friction_coefficient) == type(None):
            if type(src_file) != type(None):
                self.parameters['friction_coefficient'] = type(self.parameters['friction_coefficient'])(
                        src_file['parameters/friction_coefficient'][()])
            else:
                opt.friction_coefficient = self.parameters['friction_coefficient']
        if type(opt.fmode) == type(None):
            if type(src_file) != type(None):
                self.parameters['fmode'] = type(self.parameters['fmode'])(
                        src_file['parameters/fmode'][()])
            else:
                opt.fmode = self.parameters['fmode']
        if type(opt.famplitude) == type(None):
            if type(src_file) != type(None):
                self.parameters['famplitude'] = type(self.parameters['famplitude'])(
                        src_file['parameters/famplitude'][()])
            else:
                opt.famplitude = self.parameters['famplitude']
        if type(opt.energy) == type(None):
            if type(src_file) != type(None):
                self.parameters['energy'] = type(self.parameters['energy'])(
                        src_file['parameters/energy'][()])
            else:
                opt.energy = self.parameters['energy']
        if type(opt.fk0) == type(None):
            if type(src_file) != type(None):
                self.parameters['fk0'] = type(self.parameters['fk0'])(
                        src_file['parameters/fk0'][()])
            else:
                opt.fk0 = self.parameters['fk0']
        if type(opt.fk1) == type(None):
            if type(src_file) != type(None):
                self.parameters['fk1'] = type(self.parameters['fk1'])(
                        src_file['parameters/fk1'][()])
            else:
                opt.fk1 = self.parameters['fk1']
        if type(opt.forcing_type) == type(None):
            if type(src_file) != type(None):
                self.parameters['forcing_type'] = str(
                        src_file['parameters/forcing_type'][()].decode())
            else:
                opt.forcing_type = self.parameters['forcing_type']
        if type(opt.injection_rate) == type(None):
            if type(src_file) != type(None):
                self.parameters['injection_rate'] = type(self.parameters['injection_rate'])(
                        src_file['parameters/injection_rate'][()])
            else:
                opt.injection_rate = self.parameters['injection_rate']
        if type(opt.dealias_type) == type(None):
            if type(src_file) != type(None):
                self.parameters['dealias_type'] = type(self.parameters['dealias_type'])(
                        src_file['parameters/dealias_type'][()])
            else:
                opt.dealias_type = self.parameters['dealias_type']
        if (opt.nx > opt.n or
            opt.ny > opt.n or
            opt.nz > opt.n):
            opt.n = min(opt.nx, opt.ny, opt.nz)
            print("Warning: '-n' parameter changed to minimum of nx, ny, nz. This affects the computation of nu.")
        if type(opt.dt) == type(None):
            if self.dns_type in ['kraichnan_field']:
                opt.dt = opt.dtfactor * 0.5 / opt.n**2
            else:
                opt.dt = (opt.dtfactor / opt.n)
        self.parameters['dt'] = opt.dt
        # check value of kMax
        kM = opt.n * 0.5
        if opt.dealias_type == 1:
            kM *= 0.8

        if type(opt.nu) == type(None):
            # check if we are using viscosity of source simulation
            if type(src_file) != type(None):
                opt.nu = type(self.parameters['nu'])(
                        src_file['parameters/nu'][()])
            # empirical viscosity based on forcing type
            else:
                if opt.forcing_type == 'fixed_energy_injection_rate':
                    # use the fact that mean dissipation rate is equal to injection rate
                        opt.nu = (
                            opt.injection_rate *
                            (opt.kMeta / kM)**4)**(1./3)
                        if opt.injection_rate == 0:
                            opt.nu = (
                            (opt.kMeta / kM)**(4./3) *
                            (np.pi / 1.5)**(1./3) *
                            (2*self.parameters['energy'] / 3)**0.5)
                elif opt.forcing_type == 'fixed_energy':
                    kf = 1. / (1./opt.fk0 +
                               1./opt.fk1)
                    opt.nu = (
                        (opt.kMeta / kM)**(4./3) *
                        (np.pi / kf)**(1./3) *
                        (2*self.parameters['energy'] / 3)**0.5)
                else:
                    opt.nu = (opt.kMeta * 2 / opt.n)**(4./3)
        self.parameters['nu'] = opt.nu
        if type(opt.checkpoints_per_file) == type(None):
            # hardcoded FFTW complex representation size
            field_size = 3*(opt.nx+2)*opt.ny*opt.nz*self.fluid_dtype.itemsize
            checkpoint_size = field_size
            if self.dns_type in self.list_of_particle_codes:
                rhs_size = self.parameters['tracers0_integration_steps']
                if type(opt.tracers0_integration_steps) != type(None):
                    rhs_size = opt.tracers0_integration_steps
                nparticles = opt.nparticles
                if type(nparticles) == type(None):
                    nparticles = self.NSVEp_extra_parameters['nparticles']
                particle_size = (1+rhs_size)*3*nparticles*8
                checkpoint_size += particle_size
            if checkpoint_size < 1e9:
                opt.checkpoints_per_file = int(1e9 / checkpoint_size)
        self.pars_from_namespace(opt)
        return opt
    def launch(
            self,
            args = [],
            **kwargs):
        opt = self.prepare_launch(args = args)
        self.launch_jobs(opt = opt, **kwargs)
        return None
    def generate_tracer_state(
            self,
            rseed = None,
            species = 0,
            integration_steps = None,
            ncomponents = 3):
        try:
            if type(integration_steps) == type(None):
                integration_steps = self.NSVEp_extra_parameters['tracers0_integration_steps']
            if 'tracers{0}_integration_steps'.format(species) in self.parameters.keys():
                integration_steps = self.parameters['tracers{0}_integration_steps'.format(species)]
            if self.dns_type in [
                    'NSVEcomplex_particles',
                    'NSVE_Stokes_particles'] and species == 0:
                ncomponents = 6
            with h5py.File(self.get_checkpoint_0_fname(), 'a') as data_file:
                nn = self.parameters['nparticles']
                if not 'tracers{0}'.format(species) in data_file.keys():
                    data_file.create_group('tracers{0}'.format(species))
                    data_file.create_group('tracers{0}/rhs'.format(species))
                    data_file.create_group('tracers{0}/state'.format(species))
                data_file['tracers{0}/rhs'.format(species)].create_dataset(
                        '0',
                        shape = (integration_steps, nn, ncomponents,),
                        dtype = np.float64)
                dset = data_file['tracers{0}/state'.format(species)].create_dataset(
                        '0',
                        shape = (nn, ncomponents,),
                        dtype = np.float64)
                if not type(rseed) == type(None):
                    np.random.seed(rseed)
                cc = int(0)
                batch_size = int(1e6)
                def get_random_phases(npoints):
                    return np.random.random(
                                (npoints, 3))*2*np.pi
                def get_random_versors(npoints):
                    bla = np.random.normal(
                            size = (npoints, 3))
                    bla  /= np.sum(bla**2, axis = 1)[:, None]**.5
                    return bla
                while nn > 0:
                    if nn > batch_size:
                        dset[cc*batch_size:(cc+1)*batch_size, :3] = get_random_phases(batch_size)
                        if dset.shape[1] == 6:
                            if self.dns_type == 'NSVE_Stokes_particles':
                                dset[cc*batch_size:(cc+1)*batch_size, 3:] = self.parameters['initial_particle_vel']*get_random_versors(batch_size)
                            else:
                                dset[cc*batch_size:(cc+1)*batch_size, 3:] = get_random_versors(batch_size)
                        nn -= batch_size
                    else:
                        dset[cc*batch_size:cc*batch_size+nn, :3] = get_random_phases(nn)
                        if dset.shape[1] == 6:
                            if self.dns_type == 'NSVE_Stokes_particles':
                                dset[cc*batch_size:cc*batch_size+nn, 3:] = self.parameters['initial_particle_vel']*get_random_versors(nn)
                            else:
                                dset[cc*batch_size:cc*batch_size+nn, 3:] = get_random_versors(nn)
                        nn = 0
                    cc += 1
        except Exception as e:
            print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!WARNING')
            print('WARNING: when generating particles, caught the exception ', e)
            print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!WARNING')
        return None
    def generate_vector_field(
            self,
            rseed = 7547,
            spectra_slope = 1.,
            amplitude = 1.,
            iteration = 0,
            field_name = 'vorticity',
            write_to_file = False,
            # to switch to constant field, use generate_data_3D_uniform
            # for scalar_generator
            scalar_generator = tools.generate_data_3D):
        """generate vector field.

        The generated field is not divergence free, but it has the proper
        shape.

        :param rseed: seed for random number generator
        :param spectra_slope: spectrum of field will look like k^(-p)
        :param amplitude: all amplitudes are multiplied with this value
        :param iteration: the field is written at this iteration
        :param field_name: the name of the field being generated
        :param write_to_file: should we write the field to file?
        :param scalar_generator: which function to use for generating the
            individual components.
            Possible values: TurTLE.tools.generate_data_3D,
            TurTLE.tools.generate_data_3D_uniform
        :type rseed: int
        :type spectra_slope: float
        :type amplitude: float
        :type iteration: int
        :type field_name: str
        :type write_to_file: bool
        :type scalar_generator: function

        :returns: ``Kdata``, a complex valued 4D ``numpy.array`` that uses the
            transposed FFTW layout.
            Kdata[ky, kz, kx, i] is the amplitude of mode (kx, ky, kz) for
            the i-th component of the field.
            (i.e. x is the fastest index and z the slowest index in the
            real-space representation).
        """
        np.random.seed(rseed)
        Kdata00 = scalar_generator(
                self.parameters['nz'],
                self.parameters['ny'],
                self.parameters['nx'],
                p = spectra_slope,
                amplitude = amplitude).astype(self.ctype)
        Kdata01 = scalar_generator(
                self.parameters['nz'],
                self.parameters['ny'],
                self.parameters['nx'],
                p = spectra_slope,
                amplitude = amplitude).astype(self.ctype)
        Kdata02 = scalar_generator(
                self.parameters['nz'],
                self.parameters['ny'],
                self.parameters['nx'],
                p = spectra_slope,
                amplitude = amplitude).astype(self.ctype)
        Kdata0 = np.zeros(
                Kdata00.shape + (3,),
                Kdata00.dtype)
        Kdata0[..., 0] = Kdata00
        Kdata0[..., 1] = Kdata01
        Kdata0[..., 2] = Kdata02
        Kdata1 = tools.padd_with_zeros(
                Kdata0,
                self.parameters['nz'],
                self.parameters['ny'],
                self.parameters['nx'])
        if write_to_file:
            Kdata1.tofile(
                    os.path.join(self.work_dir,
                                 self.simname + "_c{0}_i{1:0>5x}".format(field_name, iteration)))
        return Kdata1
    def copy_complex_field(
            self,
            src_file_name,
            src_dset_name,
            dst_file,
            dst_dset_name,
            make_link = True):
        # I define a min_shape thingie, but for now I only trust this method for
        # the case of increasing/decreasing by the same factor in all directions.
        # in principle we could write something more generic, but i'm not sure
        # how complicated that would be
        dst_shape = (self.parameters['ny'],
                     self.parameters['nz'],
                     (self.parameters['nx']+2) // 2,
                     3)
        src_file = h5py.File(src_file_name, 'r')
        if (src_file[src_dset_name].shape == dst_shape):
            dst_file[dst_dset_name] = h5py.ExternalLink(
                    src_file_name,
                    src_dset_name)
        else:
            min_shape = (min(dst_shape[0], src_file[src_dset_name].shape[0]),
                         min(dst_shape[1], src_file[src_dset_name].shape[1]),
                         min(dst_shape[2], src_file[src_dset_name].shape[2]),
                         3)
            src_shape = src_file[src_dset_name].shape
            dst_file.create_dataset(
                    dst_dset_name,
                    shape = dst_shape,
                    dtype = np.dtype(self.ctype),
                    fillvalue = complex(0))
            for kz in range(min_shape[0]//2):
                dst_file[dst_dset_name][kz,:min_shape[1]//2, :min_shape[2]] = \
                        src_file[src_dset_name][kz, :min_shape[1]//2, :min_shape[2]]
                dst_file[dst_dset_name][kz,
                                        dst_shape[1] - min_shape[1]//2+1:,
                                        :min_shape[2]] = \
                        src_file[src_dset_name][kz,
                                                src_shape[1] - min_shape[1]//2+1,
                                                :min_shape[2]]
                if kz > 0:
                    dst_file[dst_dset_name][-kz,:min_shape[1]//2, :min_shape[2]] = \
                            src_file[src_dset_name][-kz, :min_shape[1]//2, :min_shape[2]]
                    dst_file[dst_dset_name][-kz,
                                            dst_shape[1] - min_shape[1]//2+1:,
                                            :min_shape[2]] = \
                            src_file[src_dset_name][-kz,
                                                    src_shape[1] - min_shape[1]//2+1,
                                                    :min_shape[2]]
        return None
    def generate_particle_data(
            self,
            opt = None):
        if (self.parameters['nparticles'] > 0):
            if (self.parameters['cpp_random_particles'] == 0):
                self.generate_tracer_state(
                        species = 0,
                        rseed = opt.particle_rand_seed)
            if not os.path.exists(self.get_particle_file_name()):
                with h5py.File(self.get_particle_file_name(), 'w') as particle_file:
                    particle_file.create_group('tracers0/position')
                    particle_file.create_group('tracers0/velocity')
                    particle_file.create_group('tracers0/acceleration')
                    if self.dns_type in ['NSVEcomplex_particles']:
                        particle_file.create_group('tracers0/orientation')
                        particle_file.create_group('tracers0/velocity_gradient')
                    if self.dns_type in ['NSVE_Stokes_particles']:
                        particle_file.create_group('tracers0/momentum')
                        particle_file.create_group('tracers0/vorticity')
                    if self.dns_type in ['NSVEp_extra_sampling']:
                        particle_file.create_group('tracers0/velocity_gradient')
                        particle_file.create_group('tracers0/pressure')
                        particle_file.create_group('tracers0/pressure_gradient')
                        particle_file.create_group('tracers0/pressure_Hessian')
        return None
    def generate_initial_condition(
            self,
            opt = None,
            need_field = None,
            checkpoint_field = None):
        """Creates the "checkpoint 0" file.

        Keyword arguments:
        opt              --- options object, this method may access any of
                                * opt.src_simname,
                                * opt.src_work_dir,
                                * opt.src_iteration
        need_field       --- boolean: does the DNS need the "0" iteration field
                             to be present in the "checkpoint 0" file?
        checkpoint_field --- string: name of field used for checkpoints, one of
                             'velocity' or 'vorticity'.
        """
        # take care of fields' initial condition
        if type(checkpoint_field) == type(None):
            if self.dns_type in ['NSE', 'NSE_alt_dealias']:
                checkpoint_field = 'velocity'
            else:
                checkpoint_field = 'vorticity'
        # first, check if initial field exists
        if type(need_field) == type(None):
            need_field = False
            if self.check_current_vorticity_exists:
                need_field = True
            if self.dns_type in [
                    'NSE',
                    'NSE_alt_dealias',
                    'NSVE',
                    'NSVE_no_output',
                    'static_field',
                    'NSVEparticles',
                    'NSVEcomplex_particles',
                    'NSVE_Stokes_particles',
                    'NSVEparticles_no_output',
                    'NSVEp_extra_sampling']:
                if not os.path.exists(self.get_checkpoint_0_fname()):
                    need_field = True
                else:
                    f = h5py.File(self.get_checkpoint_0_fname(), 'r')
                    try:
                        dset = f[checkpoint_field + '/complex/0']
                        need_field = (dset.shape != (self.parameters['ny'],
                                                     self.parameters['nz'],
                                                     self.parameters['nx']//2+1,
                                                     3))
                    except:
                        need_field = True
                    f.close()
        if need_field:
            # sanity check. User cannot demand a random initial condition and a source simulation at the same time.
            assert ((len(opt.src_simname) == 0) or (self.parameters['field_random_seed'] == 0)), \
                    'User may not request sourcing the initial condition from another simulation unless "field_random_seed" is 0. Aborting.'
            f = h5py.File(self.get_checkpoint_0_fname(), 'a')
            if len(opt.src_simname) > 0:
                source_simfile = os.path.join(
                        os.path.realpath(opt.src_work_dir),
                        opt.src_simname + '.h5')
                if os.path.exists(source_simfile):
                    source_current_checkpoint = h5py.File(
                            source_simfile, 'r')['checkpoint'][()]
                else:
                    source_current_checkpoint = 0
                source_checkpoints = [
                        os.path.join(
                            os.path.realpath(opt.src_work_dir),
                            opt.src_simname + '_checkpoint_{0}.h5'.format(cp))
                        for cp in range(source_current_checkpoint+1)]
                src_file = None
                for check_file in source_checkpoints:
                    if os.path.exists(check_file):
                        f0 = h5py.File(check_file, 'r')
                        if ('{0}'.format(opt.src_iteration)
                           in f0[checkpoint_field + '/complex'].keys()):
                            f0.close()
                            src_file = check_file
                            break
                if type(src_file) == type(None):
                    print('could not find checkpoint file')
                    print('source_checkpoints = ', source_checkpoints)
                    print('checkpoint_field = ', checkpoint_field)
                    print('iteration = ', opt.src_iteration)
                    exit(0)
                self.copy_complex_field(
                        src_file,
                        checkpoint_field + '/complex/{0}'.format(opt.src_iteration),
                        f,
                        checkpoint_field + '/complex/{0}'.format(0))
            else:
                f.require_group(checkpoint_field + '/complex')
            f.close()
        if self.dns_type == 'kraichnan_field':
            if not os.path.exists(self.get_checkpoint_0_fname()):
                f = h5py.File(self.get_checkpoint_0_fname(), 'a')
                f.create_group('velocity/real')
                f.close()
        # now take care of particles' initial condition
        if self.dns_type in self.list_of_particle_codes:
            self.generate_particle_data(opt = opt)
        return None
    def launch_jobs(
            self,
            opt = None,
            iter0 = 0):
        """Launches jobs. Prepares initial condition and writes parameters to file if necessary."""
        if (iter0 == 0):
            self.check_current_vorticity_exists = False
            self.check_current_velocity_exists = False
        else:
            if (self.dns_type in ['NSVE']):
                self.check_current_vorticity_exists = True
                self.check_current_velocity_exists = False
            if self.dns_type in ['NSE', 'NSE_alt_dealias']:
                self.check_current_vorticity_exists = False
                self.check_current_velocity_exists = True
        if not os.path.exists(self.get_data_file_name()):
            if not os.path.isdir(self.work_dir):
                os.makedirs(self.work_dir)
            self.generate_initial_condition(opt = opt)
            self.write_par(iter0 = iter0)
        self.run(
                nb_processes = opt.nb_processes,
                nb_threads_per_process = opt.nb_threads_per_process,
                njobs = opt.njobs,
                hours = opt.minutes // 60,
                minutes = opt.minutes % 60,
                no_submit = opt.no_submit,
                no_debug = not opt.debug)
        return None

