################################################################################
#                                                                              #
#  Copyright 2015-2019 Max Planck Institute for Dynamics and Self-Organization #
#                                                                              #
#  This file is part of TurTLE.                                                #
#                                                                              #
#  TurTLE is free software: you can redistribute it and/or modify              #
#  it under the terms of the GNU General Public License as published           #
#  by the Free Software Foundation, either version 3 of the License,           #
#  or (at your option) any later version.                                      #
#                                                                              #
#  TurTLE is distributed in the hope that it will be useful,                   #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with TurTLE.  If not, see <http://www.gnu.org/licenses/>              #
#                                                                              #
# Contact: Cristian.Lalescu@ds.mpg.de                                          #
#                                                                              #
################################################################################



import os
import h5py
import numpy as np

import TurTLE
from ._code import _code
from TurTLE import DNS

class TEST(_code):
    """This class is meant to stitch together the C++ code into a final source file,
    compile it, and handle all job launching.
    """
    def __init__(
            self,
            work_dir = './',
            simname = 'test'):
        _code.__init__(
                self,
                work_dir = work_dir,
                simname = simname)
        self.generate_default_parameters()
        self.check_current_vorticity_exists = False
        return None
    def write_src(self):
        self.version_message = (
                '/***********************************************************************\n' +
                '* this code automatically generated by TurTLE\n' +
                '* version {0}\n'.format(TurTLE.__version__) +
                '***********************************************************************/\n\n\n')
        self.include_list = [
                '"full_code/main_code.hpp"',
                '<cmath>',
                '"full_code/{0}.hpp"\n'.format(self.dns_type)]
        self.main = """
            int main(int argc, char *argv[])
            {{
                bool fpe = (
                    (getenv("TURTLE_FPE_OFF") == nullptr) ||
                    (getenv("TURTLE_FPE_OFF") != std::string("TRUE")));
                return main_code< {0} >(argc, argv, fpe);
            }}
            """.format(self.dns_type + '<{0}>'.format(self.C_field_dtype))
        self.includes = '\n'.join(
                ['#include ' + hh
                 for hh in self.include_list])
        with open(self.name + '.cpp', 'w') as outfile:
            outfile.write(self.version_message + '\n\n')
            outfile.write(self.includes + '\n\n')
            outfile.write(self.main + '\n')
        return None
    def generate_default_parameters(self):
        # these parameters are relevant for all TEST classes
        self.parameters['fftw_plan_rigor'] = 'FFTW_ESTIMATE'
        self.parameters['dealias_type'] = int(1)
        self.parameters['dkx'] = float(1.0)
        self.parameters['dky'] = float(1.0)
        self.parameters['dkz'] = float(1.0)
        self.parameters['filter_length'] = float(1.0)
        self.parameters['field_random_seed'] = int(1)
        return None
    def generate_extra_parameters(
            self,
            dns_type = None):
        pars = {}
        if dns_type in ['field_output_test']:
            pars['number_operations'] = int(1)
        if dns_type == 'test_interpolation':
            pars['nparticles'] = 3
            pars['tracers0_integration_steps'] = int(4)
            pars['tracers0_neighbours'] = int(1)
            pars['tracers0_smoothness'] = int(1)
        if dns_type == 'test_interpolation_methods':
            pars['nxparticles'] = 160
            pars['nzparticles'] = 180
        if dns_type == 'test_particle_integration':
            pars['nxparticles'] = 13
            pars['nzparticles'] = 11
            pars['dt'] = 0.125
        if dns_type == 'phase_shift_test':
            pars['random_phase_seed'] = 1
        if dns_type in ['dealias_test', 'Gauss_field_test']:
            pars['output_incompressible_field'] = int(0)
            pars['histogram_bins'] = int(129)
            pars['max_velocity_estimate'] = float(8)
            pars['spectrum_dissipation'] = float(0.027)
            pars['spectrum_Lint'] = float(1.3)
            pars['spectrum_etaK'] = float(0.3)
            pars['spectrum_large_scale_const'] = float(6.78)
            pars['spectrum_small_scale_const'] = float(0.40)
        return pars
    def get_kspace(self):
        kspace = {}
        if self.parameters['dealias_type'] == 1:
            kMx = self.parameters['dkx']*(self.parameters['nx']//2 - 1)
            kMy = self.parameters['dky']*(self.parameters['ny']//2 - 1)
            kMz = self.parameters['dkz']*(self.parameters['nz']//2 - 1)
        else:
            kMx = self.parameters['dkx']*(self.parameters['nx']//3 - 1)
            kMy = self.parameters['dky']*(self.parameters['ny']//3 - 1)
            kMz = self.parameters['dkz']*(self.parameters['nz']//3 - 1)
        kspace['kM'] = max(kMx, kMy, kMz)
        kspace['dk'] = min(self.parameters['dkx'],
                           self.parameters['dky'],
                           self.parameters['dkz'])
        nshells = int(kspace['kM'] / kspace['dk']) + 2
        kspace['nshell'] = np.zeros(nshells, dtype = np.int64)
        kspace['kshell'] = np.zeros(nshells, dtype = np.float64)
        kspace['kx'] = np.arange( 0,
                                  self.parameters['nx']//2 + 1).astype(np.float64)*self.parameters['dkx']
        kspace['ky'] = np.arange(-self.parameters['ny']//2 + 1,
                                  self.parameters['ny']//2 + 1).astype(np.float64)*self.parameters['dky']
        kspace['ky'] = np.roll(kspace['ky'], self.parameters['ny']//2+1)
        kspace['kz'] = np.arange(-self.parameters['nz']//2 + 1,
                                  self.parameters['nz']//2 + 1).astype(np.float64)*self.parameters['dkz']
        kspace['kz'] = np.roll(kspace['kz'], self.parameters['nz']//2+1)
        return kspace
    def write_par(
            self,
            iter0 = 0,
            particle_ic = None):
        assert (iter0 == 0)
        _code.write_par(self, iter0 = iter0)
        with h5py.File(self.get_data_file_name(), 'r+') as ofile:
            ofile['code_info/exec_name'] = self.name
            kspace = self.get_kspace()
            for k in kspace.keys():
                ofile['kspace/' + k] = kspace[k]
            nshells = kspace['nshell'].shape[0]
            kspace = self.get_kspace()
            nshells = kspace['nshell'].shape[0]
            ofile['checkpoint'] = int(0)
            vec_spectra_stats = []
            tens_rspace_stats = []
            vec4_rspace_stats = []
            scal_rspace_stats = []
            if self.dns_type in ['dealias_test']:
                vec_spectra_stats.append('field3')
                vec_spectra_stats.append('field4')
                vec4_rspace_stats.append('field3')
                vec4_rspace_stats.append('field4')
            if self.dns_type in ['Gauss_field_test']:
                vec_spectra_stats.append('velocity')
                vec_spectra_stats.append('k*velocity')
                vec_spectra_stats.append('k2*velocity')
                vec4_rspace_stats.append('velocity')
                tens_rspace_stats.append('velocity_gradient')
                scal_rspace_stats.append('velocity_divergence')
            for k in vec_spectra_stats:
                time_chunk = 2**20//(8*3*3*nshells)
                time_chunk = max(time_chunk, 1)
                ofile.create_dataset('statistics/spectra/' + k + '_' + k,
                                     (1, nshells, 3, 3),
                                     chunks = (time_chunk, nshells, 3, 3),
                                     maxshape = (None, nshells, 3, 3),
                                     dtype = np.float64)
            for k in scal_rspace_stats:
                time_chunk = 2**20//(8*10)
                time_chunk = max(time_chunk, 1)
                a = ofile.create_dataset('statistics/moments/' + k,
                                     (1, 10),
                                     chunks = (time_chunk, 10),
                                     maxshape = (None, 10),
                                     dtype = np.float64)
                time_chunk = 2**20//(8*self.parameters['histogram_bins'])
                time_chunk = max(time_chunk, 1)
                ofile.create_dataset('statistics/histograms/' + k,
                                     (1,
                                      self.parameters['histogram_bins']),
                                     chunks = (time_chunk,
                                               self.parameters['histogram_bins']),
                                     maxshape = (None,
                                                 self.parameters['histogram_bins']),
                                     dtype = np.int64)
            for k in vec4_rspace_stats:
                time_chunk = 2**20//(8*4*10)
                time_chunk = max(time_chunk, 1)
                a = ofile.create_dataset('statistics/moments/' + k,
                                     (1, 10, 4),
                                     chunks = (time_chunk, 10, 4),
                                     maxshape = (None, 10, 4),
                                     dtype = np.float64)
                time_chunk = 2**20//(8*4*self.parameters['histogram_bins'])
                time_chunk = max(time_chunk, 1)
                ofile.create_dataset('statistics/histograms/' + k,
                                     (1,
                                      self.parameters['histogram_bins'],
                                      4),
                                     chunks = (time_chunk,
                                               self.parameters['histogram_bins'],
                                               4),
                                     maxshape = (None,
                                                 self.parameters['histogram_bins'],
                                                 4),
                                     dtype = np.int64)
            for k in tens_rspace_stats:
                time_chunk = 2**20//(8*9*10)
                time_chunk = max(time_chunk, 1)
                a = ofile.create_dataset('statistics/moments/' + k,
                                     (1, 10, 3, 3),
                                     chunks = (time_chunk, 10, 3, 3),
                                     maxshape = (None, 10, 3, 3),
                                     dtype = np.float64)
                time_chunk = 2**20//(8*9*self.parameters['histogram_bins'])
                time_chunk = max(time_chunk, 1)
                ofile.create_dataset('statistics/histograms/' + k,
                                     (1,
                                      self.parameters['histogram_bins'],
                                      3, 3),
                                     chunks = (time_chunk,
                                               self.parameters['histogram_bins'],
                                               3, 3),
                                     maxshape = (None,
                                                 self.parameters['histogram_bins'],
                                                 3, 3),
                                     dtype = np.int64)
        return None
    def simulation_parser_arguments(
            self,
            parser):
        parser.add_argument(
                '--simname',
                type = str, dest = 'simname',
                default = 'test')
        parser.add_argument(
               '-n', '--grid-size',
               type = int,
               dest = 'n',
               default = 32,
               metavar = 'N',
               help = 'code is run by default in a grid of NxNxN')
        for coord in ['x', 'y', 'z']:
            parser.add_argument(
                   '--L{0}'.format(coord), '--box-length-{0}'.format(coord),
                   type = float,
                   dest = 'L{0}'.format(coord),
                   default = 2.0,
                   metavar = 'length{0}'.format(coord),
                   help = 'length of the box in the {0} direction will be `length{0} x pi`'.format(coord))
        parser.add_argument(
                '--wd',
                type = str, dest = 'work_dir',
                default = './')
        parser.add_argument(
                '--precision',
                choices = ['single', 'double'],
                type = str,
                default = 'single')
        return None
    def add_parser_arguments(
            self,
            parser):
        subparsers = parser.add_subparsers(
                dest = 'TEST_class',
                help = 'type of simulation to run')
        subparsers.required = True
        parser_dealias_test = subparsers.add_parser(
                'dealias_test',
                help = 'generate spectra for two thirds and one half dealiasing')
        parser_phase_shift_test = subparsers.add_parser(
                'phase_shift_test',
                help = 'test phase shift functionality')
        parser_Gauss_field_test = subparsers.add_parser(
                'Gauss_field_test',
                help = 'test incompressible Gaussian random fields')
        parser_filter_test = subparsers.add_parser(
                'filter_test',
                help = 'plain filter test')
        parser_write_filtered_test = subparsers.add_parser(
                'write_filtered_test',
                help = 'test of write_filtered functionality')
        parser_shared_array_test = subparsers.add_parser(
                'shared_array_test',
                help = 'plain shared array test')
        parser_field_test = subparsers.add_parser(
                'field_test',
                help = 'plain field test')
        parser_symmetrize_test = subparsers.add_parser(
                'symmetrize_test',
                help = 'plain symmetrize test')
        parser_field_output_test = subparsers.add_parser(
                'field_output_test',
                help = 'plain field output test')
        parser_test_interpolation = subparsers.add_parser(
                'test_interpolation',
                help = 'test velocity gradient interpolation')
        parser_test_interpolation_methods = subparsers.add_parser(
                'test_interpolation_methods',
                help = 'test interpolation methods')
        parser_test_particle_integration = subparsers.add_parser(
                'test_particle_integration',
                help = 'test interpolation methods')
        parser_test_tracer_set = subparsers.add_parser(
                'test_tracer_set',
                help = 'basic test of tracer set solver')

        for parser in ['dealias_test',
                       'phase_shift_test',
                       'Gauss_field_test',
                       'filter_test',
                       'write_filtered_test',
                       'shared_array_test',
                       'field_test',
                       'symmetrize_test',
                       'field_output_test',
                       'test_interpolation',
                       'test_interpolation_methods',
                       'test_particle_integration',
                       'test_tracer_set']:
            eval('self.simulation_parser_arguments(parser_' + parser + ')')
            eval('self.job_parser_arguments(parser_' + parser + ')')
            eval('self.parameters_to_parser_arguments(parser_' + parser + ')')
            eval('self.parameters_to_parser_arguments(parser_' + parser + ', self.generate_extra_parameters(dns_type = \'' + parser + '\'))')
        return None
    def prepare_launch(
            self,
            args = []):
        opt = _code.prepare_launch(self, args = args)
        self.set_precision(opt.precision)
        self.dns_type = opt.TEST_class
        self.name = self.dns_type + '-' + self.fluid_precision + '-v' + TurTLE.__version__
        self.parameters.update(
                self.generate_extra_parameters(dns_type = self.dns_type))
        # merge parameters if needed
        self.pars_from_namespace(opt)
        return opt
    def launch(
            self,
            args = [],
            **kwargs):
        opt = self.prepare_launch(args = args)
        self.launch_jobs(opt = opt, **kwargs)
        return None
    def launch_jobs(
            self,
            opt = None,
            particle_initial_condition = None):
        if not os.path.exists(os.path.join(self.work_dir, self.simname + '.h5')):
            # the write_par call creates the work dir fi needed
            self.write_par(
                    particle_ic = particle_initial_condition)
            if self.dns_type == 'test_tracer_set':
                ofile = h5py.File(os.path.join(self.work_dir, 'test_particle_sample.h5'), 'w')
                ofile.create_group('tracers0')
                ofile.create_group('tracers0/position')
                ofile.create_group('tracers0/velocity')
                ofile.close()
            if self.dns_type == 'test_interpolation_methods':
                ofile = h5py.File(os.path.join(self.work_dir, self.simname + '_particles.h5'), 'w')
                ofile.create_group('particle')
                ofile.create_group('particle/position')
                for n, m in [(0, 0),
                             (1, 1),
                             (2, 1), (2, 2),
                             (3, 1), (3, 2),
                             (4, 1), (4, 2),
                             (5, 1), (5, 2)]:
                    ofile.create_group('particle/phi_{0}{1}'.format(n, m))
                ofile.close()
            if self.dns_type == 'test_particle_integration':
                ofile = h5py.File(os.path.join(self.work_dir, self.simname + '_particles.h5'), 'w')
                ofile.create_group('particle')
                ofile.create_group('particle/position')
                for n, m in [(0, 0),
                             (1, 1),
                             (2, 1), (2, 2),
                             (3, 1), (3, 2),
                             (4, 1), (4, 2),
                             (5, 1), (5, 2)]:
                    for integration_order in [1, 2, 4]:
                        for factor in [1, 2, 4, 8]:
                            ofile.create_group(
                                    'particle_o{0}_n{1}_m{2}_f{3}/position'.format(
                                        integration_order, n, m, factor))
                            ofile.create_group(
                                    'particle_o{0}_n{1}_m{2}_f{3}/velocity'.format(
                                        integration_order, n, m, factor))
                ofile.close()
            if self.dns_type == 'test_interpolation':
                if type(particle_initial_condition) == type(None):
                    pbase_shape = (self.parameters['nparticles'],)
                    number_of_particles = self.parameters['nparticles']
                else:
                    pbase_shape = particle_initial_condition.shape[:-1]
                    assert(particle_initial_condition.shape[-1] == 3)
                    number_of_particles = 1
                    for val in pbase_shape[1:]:
                        number_of_particles *= val
                ncomponents = 3
                with h5py.File(os.path.join(self.work_dir, self.simname + '_input.h5'), 'a') as ofile:
                    s = 0
                    ofile.create_group('tracers{0}'.format(s))
                    ofile.create_group('tracers{0}/rhs'.format(s))
                    ofile.create_group('tracers{0}/state'.format(s))
                    ofile['tracers{0}/rhs'.format(s)].create_dataset(
                            '0',
                            shape = (
                                (self.parameters['tracers{0}_integration_steps'.format(s)],) +
                                pbase_shape +
                                (ncomponents,)),
                            dtype = np.float64)
                    ofile['tracers{0}/state'.format(s)].create_dataset(
                            '0',
                            shape = (
                                pbase_shape +
                                (ncomponents,)),
                            dtype = np.float64)
                    if type(particle_initial_condition) == type(None):
                        ofile['tracers0/state/0'][:] = np.random.random(pbase_shape + (ncomponents,))*2*np.pi
                    else:
                        ofile['tracers0/state/0'][:] = particle_initial_condition
                with h5py.File(os.path.join(self.work_dir, self.simname + '_input.h5'), 'a') as ofile:
                    data = DNS.generate_vector_field(self,
                           write_to_file = False,
                           spectra_slope = 1.0,
                           amplitude = 0.05)
                    #data[:] = 0.0
                    ## ABC
                    #data[0, 0, 1, 1] = -0.5*(1j)
                    #data[0, 0, 1, 2] =  0.5*(1j)
                    #data[0, 1, 0, 0]                         = -0.5*(1j)
                    #data[0, self.parameters['nz'] - 1, 0, 0] =  0.5*(1j)
                    #data[0, 1, 0, 1]                         =  0.5
                    #data[0, self.parameters['nz'] - 1, 0, 1] =  0.5
                    #data[1, 0, 0, 0]                         =  0.5
                    #data[self.parameters['ny'] - 1, 0, 0, 0] =  0.5
                    #data[1, 0, 0, 2]                         = -0.5*(1j)
                    #data[self.parameters['ny'] - 1, 0, 0, 2] =  0.5*(1j)
                    ofile['vorticity/complex/{0}'.format(0)] = data
                with h5py.File(os.path.join(self.work_dir, self.simname + '_output.h5'), 'a') as ofile:
                    ofile.require_group('tracers0')
                    for kk in ['position', 'velocity', 'vorticity', 'velocity_gradient']:
                        ofile['tracers0'].require_group(kk)


        self.run(
                nb_processes = opt.nb_processes,
                nb_threads_per_process = opt.nb_threads_per_process,
                njobs = opt.njobs,
                hours = opt.minutes // 60,
                minutes = opt.minutes % 60,
                no_submit = opt.no_submit)
        return None
