################################################################################
#                                                                              #
#  Copyright 2015-2019 Max Planck Institute for Dynamics and Self-Organization #
#                                                                              #
#  This file is part of TurTLE.                                                #
#                                                                              #
#  TurTLE is free software: you can redistribute it and/or modify              #
#  it under the terms of the GNU General Public License as published           #
#  by the Free Software Foundation, either version 3 of the License,           #
#  or (at your option) any later version.                                      #
#                                                                              #
#  TurTLE is distributed in the hope that it will be useful,                   #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
#  GNU General Public License for more details.                                #
#                                                                              #
#  You should have received a copy of the GNU General Public License           #
#  along with TurTLE.  If not, see <http://www.gnu.org/licenses/>              #
#                                                                              #
# Contact: Cristian.Lalescu@ds.mpg.de                                          #
#                                                                              #
################################################################################



import os
import sys
import numpy as np
import h5py
from TurTLE import __version__

class _base(object):
    """This class contains simulation parameters, and handles parameter related
    functionalities of both python objects and C++ codes.
    """
    def __init__(
            self,
            work_dir = './',
            simname = 'test'):
        self.iorank = 0
        ### simulation parameters
        self.parameters = {'nx' : 32,
                           'ny' : 32,
                           'nz' : 32,
                           'niter_todo' : int(8),
                           'niter_stat' : int(1),
                           'niter_out' : int(8),
                           'checkpoints_per_file' : int(1),
                           }
        self.parameter_description = {
                'nx'         : 'Number of real-space grid nodes in the x direction (fastest array index in real-space representation).',
                'ny'         : 'Number of real-space grid nodes in the y direction.',
                'nz'         : 'Number of real-space grid nodes in the z direction (slowest array index in real-space representation).',
                'niter_todo' : 'Number of iterations to compute during a single run.',
                'niter_stat' : 'Interval (in iterations) over which to compute field statistics (i.e. call `do_stats`).',
                'niter_out'  : 'Output is performed every `NITER_OUT` iterations.',
                'checkpoints_per_file' : 'Number of checkpoints to store in a single checkpoint file. Rule of thumb: files should hold gigabytes of data, rather than megabytes.',
                }
        self.string_length = 512
        self.work_dir = os.path.realpath(work_dir)
        self.simname = simname
        return None
    def cdef_pars(
            self,
            parameters = None):
        if type(parameters) == type(None):
            parameters = self.parameters
        key = sorted(list(parameters.keys()))
        src_txt = ''
        for i in range(len(key)):
            if (type(parameters[key[i]]) == int  and parameters[key[i]] >= 1<<30):
                src_txt += 'long long int ' + key[i] + ';\n'
            elif type(parameters[key[i]]) == int:
                src_txt += 'int ' + key[i] + ';\n'
            elif type(parameters[key[i]]) == str:
                src_txt += 'char ' + key[i] + '[{0}];\n'.format(self.string_length)
            elif type(parameters[key[i]]) == np.ndarray:
                src_txt += 'std::vector<'
                if parameters[key[i]].dtype == np.float64:
                    src_txt += 'double'
                elif parameters[key[i]].dtype == int:
                    src_txt += 'int'
                src_txt += '> ' + key[i] + ';\n'
            else:
                src_txt += 'double ' + key[i] + ';\n'
        return src_txt
    def cread_pars(
            self,
            parameters = None,
            function_suffix = '',
            template_class = '',
            template_prefix = '',
            file_group = 'parameters',
            just_declaration = False,
            simname_variable = 'simname',
            prepend_this = False):
        if type(parameters) == type(None):
            parameters = self.parameters
        key = sorted(list(parameters.keys()))
        src_txt = (template_prefix +
                   'int ' +
                   template_class +
                   'read_parameters' + function_suffix + '()')
        if just_declaration:
            return src_txt + ';\n'
        src_txt += ('\n{\n' +
                    'hid_t parameter_file;\n' +
                    'hid_t dset, memtype, space;\n' +
                    'char fname[256];\n' +
                    'hsize_t dims[1];\n' +
                    'char *string_data;\n' +
                    'snprintf(fname, 255, "%s.h5", {0});\n'.format(simname_variable) +
                    'parameter_file = H5Fopen(fname, H5F_ACC_RDONLY, H5P_DEFAULT);\n')
        key_prefix = ''
        if prepend_this:
            key_prefix = 'this->'
        for i in range(len(key)):
            src_txt += 'dset = H5Dopen(parameter_file, "/{0}/{1}", H5P_DEFAULT);\n'.format(
                    file_group, key[i])
            if (type(parameters[key[i]]) == int and parameters[key[i]] >= 1<<30):
                src_txt += ('if (dset > 0) H5Dread(dset, H5T_NATIVE_LLONG, H5S_ALL, H5S_ALL, H5P_DEFAULT, &{0});\n'
                          + 'else {0} = 0;\n').format(key_prefix + key[i])
            elif type(parameters[key[i]]) == int:
                src_txt += ('if (dset > 0) H5Dread(dset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, &{0});\n'
                          + 'else {0} = 0;\n').format(key_prefix + key[i])
            elif type(parameters[key[i]]) == str:
                src_txt += ('if (dset > 0)\n' +
                            '{\n'
                            'space = H5Dget_space(dset);\n' +
                            'memtype = H5Dget_type(dset);\n' +
                            'string_data = (char*)malloc(256);\n' +
                            'H5Dread(dset, memtype, H5S_ALL, H5S_ALL, H5P_DEFAULT, &string_data);\n' +
                            'snprintf({0}, 255, "%s", string_data);\n'.format(key_prefix + key[i]) +
                            'free(string_data);\n'
                            'H5Sclose(space);\n' +
                            'H5Tclose(memtype);\n' +
                            '}\n' +
                            'else printf({0}, "NULL");\n'.format(key_prefix + key[i]))
            elif type(parameters[key[i]]) == np.ndarray:
                if parameters[key[i]].dtype in [int, np.int64, np.int32]:
                    template_par = 'int'
                elif parameters[key[i]].dtype == np.float64:
                    template_par = 'double'
                src_txt += '{0} = hdf5_tools::read_vector<{1}>(parameter_file, "/{2}/{0}");\n'.format(
                        key_prefix + key[i], template_par, file_group)
            else:
                src_txt += ('if (dset > 0) H5Dread(dset, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, &{0});\n' +
                            'else {0} = 0.0;\n').format(key_prefix + key[i])
            src_txt += 'H5Dclose(dset);\n'
        src_txt += 'H5Fclose(parameter_file);\n'
        src_txt += 'return EXIT_SUCCESS;\n}\n' # finishing read_parameters
        return src_txt
    def cprint_pars(self):
        key = sorted(list(self.parameters.keys()))
        src_txt = ''
        for i in range(len(key)):
            if type(self.parameters[key[i]]) == int:
                src_txt += 'DEBUG_MSG("'+ key[i] + ' = %d\\n", ' + key[i] + ');\n'
            elif type(self.parameters[key[i]]) == str:
                src_txt += 'DEBUG_MSG("'+ key[i] + ' = %s\\n", ' + key[i] + ');\n'
            elif type(self.parameters[key[i]]) == np.ndarray:
                src_txt += ('for (unsigned int array_counter=0; array_counter<' +
                            key[i] +
                            '.size(); array_counter++)\n' +
                            '{\n' +
                            'DEBUG_MSG("' + key[i] + '[%d] = %')
                if self.parameters[key[i]].dtype == int:
                    src_txt += 'd'
                elif self.parameters[key[i]].dtype == np.float64:
                    src_txt += 'g'
                src_txt += ('\\n", array_counter, ' +
                            key[i] +
                            '[array_counter]);\n}\n')
            else:
                src_txt += 'DEBUG_MSG("'+ key[i] + ' = %g\\n", ' + key[i] + ');\n'
        return src_txt
    def write_par(
            self,
            iter0 = 0,
            force = False):
        if not os.path.isdir(self.work_dir):
            os.makedirs(self.work_dir)
        if not os.path.exists(os.path.join(self.work_dir, self.simname + '.h5')) or force:
            ofile = h5py.File(os.path.join(self.work_dir, self.simname + '.h5'), 'w-')
            for k in self.parameters.keys():
                if (type(self.parameters[k]) == str) and (sys.version_info[0] == 3):
                    ofile['parameters/' + k] = bytes(self.parameters[k], 'ascii')
                else:
                    ofile['parameters/' + k] = self.parameters[k]
            ofile['iteration'] = int(iter0)
            ofile['code_info/solver_class'] = type(self).__name__
            ofile['code_info/VERSION'] = __version__
            return ofile.close()
        else:
            print('Warning: you are attempting to overwrite simulation parameters, but did not specify "force = True". I will not overwrite --- you should be using "rewrite", in any case.')
        return None
    def rewrite_par(
            self,
            group = None,
            parameters = None,
            file_name = None):
        assert(group != 'parameters')
        if type(file_name) == type(None):
            file_name = os.path.join(self.work_dir, self.simname + '.h5')
        ofile = h5py.File(file_name, 'a')
        for k in parameters.keys():
            if group not in ofile.keys():
                ofile.create_group(group)
            if k not in ofile[group].keys():
                if (type(parameters[k]) == str) and (sys.version_info[0] == 3):
                    ofile[group + '/' + k] = bytes(parameters[k], 'ascii')
                else:
                    ## this code to be used when h5py is fixed
                    # apparently things do work as expected when
                    # the "chunks" parameters is omitted. however,
                    # the default chunk size is the initial size of
                    # the dataset, which could be disastrous for performance...
                    #if (type(parameters[k]) == np.ndarray):
                    #    #ofile[group].create_dataset(
                    #    #        k,
                    #    #        parameters[k].shape,
                    #    #        chunks = (128,),
                    #    #        maxshape=(None),
                    #    #        dtype = parameters[k].dtype)
                    #    ofile[group + '/' + k][...] = parameters[k]
                    #else:
                    ofile[group + '/' + k] = parameters[k]
            else:
                if (type(parameters[k]) == str) and (sys.version_info[0] == 3):
                    ofile[group + '/' + k][...] = bytes(parameters[k], 'ascii')
                else:
                    if (type(parameters[k]) == np.ndarray):
                        if ofile[group + '/' + k].shape[0] != parameters[k].shape[0]:
                            del ofile[group + '/' + k]
                            ofile[group + '/' + k] = parameters[k]
                            #ofile[group + '/' + k].resize(parameters[k].shape)
                        else:
                            ofile[group + '/' + k][...] = parameters[k]
                    ofile[group + '/' + k][...] = parameters[k]
        ofile.close()
        return None
    def read_parameters(self, fname = None):
        if type(fname) == type(None):
            fname = os.path.join(self.work_dir, self.simname + '.h5')
        with h5py.File(fname, 'r') as data_file:
            for k in data_file['parameters'].keys():
                if k in self.parameters.keys():
                    if type(self.parameters[k]) in [int, float]:
                        self.parameters[k] = type(self.parameters[k])(data_file['parameters/' + k][...])
                        if data_file['parameters/' + k].size == 1:
                            self.parameters[k] = type(self.parameters[k])(data_file['parameters/' + k][()])
                        else:
                            self.parameters[k] = np.array(data_file['parameters/' + k][()]).astype(type(self.parameters[k]))
                    elif type(self.parameters[k]) == str:
                        assert(type(data_file['parameters/' + k][()]) == bytes)
                        self.parameters[k] = data_file['parameters/' + k][()].decode('ascii')
                    else:
                        self.parameters[k] = data_file['parameters/' + k][()]
                else:
                    values = data_file['parameters/' + k][()]
                    if type(values) == bytes:
                        self.parameters[k] = str(values.decode('ascii'))
                    else:
                        if values.size == 1:
                            self.parameters[k] = values.dtype.type(values)
                        else:
                            self.parameters[k] = values
        return None
    def pars_from_namespace(
            self,
            opt,
            parameters = None,
            get_sim_info = True):
        if type(parameters) == type(None):
            parameters = self.parameters
        cmd_line_pars = vars(opt)
        if 'n' in cmd_line_pars.keys():
            for k in ['nx', 'ny', 'nz']:
                if type(cmd_line_pars[k]) == type(None):
                    cmd_line_pars[k] = opt.n
        for k in parameters.keys():
            if k in cmd_line_pars.keys():
                if not type(cmd_line_pars[k]) == type(None):
                    parameters[k] = cmd_line_pars[k]
        if get_sim_info:
            self.simname = opt.simname
            self.work_dir = os.path.realpath(opt.work_dir)
        return None
    def get_coord(self, direction):
        assert(direction == 'x' or direction == 'y' or direction == 'z')
        return np.arange(.0, self.parameters['n' + direction])*2*np.pi / self.parameters['n' + direction]
    def add_parser_arguments(
            self,
            parser):
        self.job_parser_arguments(parser)
        self.parameters_to_parser_arguments(parser)
        return None
    def job_parser_arguments(
            self,
            parser):
        parser.add_argument(
                '-v', '--version',
                action = 'version',
                version = '%(prog)s ' + __version__)
        parser.add_argument(
                '--ncpu',
                type = int,
                dest = 'ncpu',
                default = -1)
        parser.add_argument(
                '--np', '--nprocesses',
                metavar = 'NPROCESSES',
                help = 'number of mpi processes to use',
                type = int,
                dest = 'nb_processes',
                default = 4)
        parser.add_argument(
                '--ntpp', '--nthreads-per-process',
                type = int,
                dest = 'nb_threads_per_process',
                metavar = 'NTHREADS_PER_PROCESS',
                help = 'number of threads to use per MPI process',
                default = 1)
        parser.add_argument(
                '--no-submit',
                action = 'store_true',
                dest = 'no_submit')
        parser.add_argument(
                '--debug',
                action = 'store_true',
                dest = 'debug')
        parser.add_argument(
                '--environment',
                type = str,
                dest = 'environment',
                default = None)
        parser.add_argument(
                '--minutes',
                type = int,
                dest = 'minutes',
                default = 5,
                help = 'If environment supports it, this is the requested wall-clock-limit.')
        parser.add_argument(
               '--njobs',
               type = int, dest = 'njobs',
               default = 1)
        parser.add_argument(
                '--profile-with-vtune',
                action = 'store_true',
                dest = 'use_vtune')
        parser.add_argument(
                '--profile-with-aps',
                action = 'store_true',
                dest = 'use_aps')
        return None
    def parameters_to_parser_arguments(
            self,
            parser,
            parameters = None,
            parameter_description = None):
        """Adds dict of parameters to parser arguments."""
        if type(parameters) == type(None):
            parameters = self.parameters
        if type(parameter_description) == type(None):
            parameter_description = self.parameter_description
        for k in sorted(parameters.keys()):
            if k in parameter_description.keys():
                description = parameter_description[k]
            else:
                description = 'No description available.'
            parser.add_argument(
                    '--{0}'.format(k),
                    type = type(parameters[k]),
                    dest = k,
                    help = description,
                    default = None)
        return None
    def get_data_file_name(self):
        return os.path.join(self.work_dir, self.simname + '.h5')
    def get_data_file(self):
        return h5py.File(self.get_data_file_name(), 'r')
    def get_particle_file_name(self):
        return os.path.join(self.work_dir, self.simname + '_particles.h5')
    def get_particle_file(self):
        return h5py.File(self.get_particle_file_name(), 'r')
    def get_cache_file_name(self):
        return os.path.join(self.work_dir, self.simname + '_cache.h5')
    def get_cache_file(self):
        return h5py.File(self.get_cache_file_name(), 'r')
    def get_postprocess_file_name(self):
        return self.get_cache_file_name()
    def get_postprocess_file(self):
        return h5py.File(self.get_postprocess_file_name(), 'r')
    def get_checkpoint_0_fname(self):
        return os.path.join(
                    self.work_dir,
                    self.simname + '_checkpoint_0.h5')
    def get_checkpoint_fname(self, iteration = 0, checkpoint = None):
        if type(checkpoint) == type(None):
            checkpoint = (iteration // self.parameters['niter_out']) // self.parameters['checkpoints_per_file']
        return os.path.join(
                    self.work_dir,
                    self.simname + '_checkpoint_{0}.h5'.format(checkpoint))
    def get_fields_fname(self):
        return os.path.join(self.work_dir, self.simname + '_fields.h5')

