import os
import sys
import argparse
import getpass

import numpy as np
import h5py

import TurTLE
import TurTLE._base
import TurTLE.DNS
from TurTLE._code import _code

cpp_location = os.path.join(
        TurTLE.data_dir, 'profiler')


class ADNS(TurTLE.DNS):
    def __init__(
            self,
            dns_type = 'scalar_evolution',
            **kwargs):
        self.dns_type = dns_type
        TurTLE.DNS.__init__(self, **kwargs)
        return None
    def write_src(
            self):
        self.version_message = (
                '/***********************************************************************\n' +
                '* this code automatically generated by TurTLE\n' +
                '* version {0}\n'.format(TurTLE.__version__) +
                '***********************************************************************/\n\n\n')
        self.include_list = [
                '"base.hpp"',
                '"scope_timer.hpp"',
                '"fftw_interface.hpp"',
                '"full_code/main_code.hpp"',
                '"full_code/NSVEparticles.hpp"',
                '<cmath>',
                '<iostream>',
                '<hdf5.h>',
                '<string>',
                '<cstring>',
                '<fftw3-mpi.h>',
                '<omp.h>',
                '<cfenv>',
                '<cstdlib>']
        self.main = """
            int main(int argc, char *argv[])
            {{
                bool fpe = (
                    (getenv("BFPS_FPE_OFF") == nullptr) ||
                    (getenv("BFPS_FPE_OFF") != std::string("TRUE")));
                return main_code< {0} >(argc, argv, fpe);
            }}
            """.format(self.dns_type + '<{0}>'.format(self.C_field_dtype))
        self.includes = '\n'.join(
                ['#include ' + hh
                 for hh in self.include_list])
        self.name = 'scalar_evolution'
        self.definitions += open(
            os.path.join(
                cpp_location, self.dns_type + '.hpp'), 'r').read()
        self.definitions += open(
            os.path.join(
                cpp_location, self.dns_type + '.cpp'), 'r').read()
        with open(self.name + '.cpp', 'w') as outfile:
            outfile.write(self.version_message + '\n\n')
            outfile.write(self.includes + '\n\n')
            outfile.write(self.definitions + '\n\n')
            outfile.write(self.main + '\n')
        self.check_current_vorticity_exists = True
        return None
    def generate_default_parameters(self):
        # these parameters are relevant for all DNS classes
        self.parameters['fftw_plan_rigor'] = 'FFTW_ESTIMATE'
        self.parameters['dealias_type'] = int(1)
        self.parameters['dkx'] = float(1.0)
        self.parameters['dky'] = float(1.0)
        self.parameters['dkz'] = float(1.0)
        self.parameters['niter_todo'] = int(12)
        self.parameters['niter_stat'] = int(4)
        self.parameters['niter_out'] = int(12)
        self.parameters['checkpoints_per_file'] = int(1)
        self.parameters['dt'] = float(0.0001)
        self.parameters['histogram_bins'] = int(64)
        self.parameters['max_value_estimate'] = float(1)
        self.parameters['field_random_seed'] = int(1)

        self.parameters['nu'] = float(0.1)
        if self.dns_type == 'scalar_evolution':
            self.parameters['mu'] = float(0.2)
            self.parameters['injection_rate'] = float(0.4)
            self.parameters['fk0'] = float(2.0)
            self.parameters['fk1'] = float(4.0)

        self.parameters['spectrum_dissipation'] = float(0.027)
        self.parameters['spectrum_Lint'] = float(0.5)
        self.parameters['spectrum_etaK'] = float(3.0)
        self.parameters['spectrum_large_scale_const'] = float(6.78)
        self.parameters['spectrum_small_scale_const'] = float(0.40)
        return None
    def prepare_launch(
            self,
            args = [],
            **kwargs):
        parser = argparse.ArgumentParser('turtle ' + type(self).__name__)
        self.job_parser_arguments(parser)
        self.simulation_parser_arguments(parser)
        self.parameters_to_parser_arguments(parser)
        opt = parser.parse_args(args)
        self.simname=opt.simname
        self.set_precision(opt.precision)
        self.name = self.dns_type + '-' + self.fluid_precision + '-v' + TurTLE.__version__
        if ((self.parameters['niter_todo'] % self.parameters['niter_out']) != 0):
            self.parameters['niter_out'] = self.parameters['niter_todo']
        if len(opt.src_work_dir) == 0:
            opt.src_work_dir = os.path.realpath(opt.work_dir)
        if type(opt.dkx) == type(None):
            opt.dkx = 2. / opt.Lx
        if type(opt.dky) == type(None):
            opt.dky = 2. / opt.Ly
        if type(opt.dkz) == type(None):
            opt.dkz = 2. / opt.Lz
        if type(opt.nx) == type(None):
            opt.nx = opt.n
        if type(opt.ny) == type(None):
            opt.ny = opt.n
        if type(opt.nz) == type(None):
            opt.nz = opt.n
        if self.dns_type == 'scalar_evolution':
            if type(opt.fk0) == type(None):
                opt.fk0 = self.parameters['fk0']
            if type(opt.fk1) == type(None):
                opt.fk1 = self.parameters['fk1']
            if type(opt.injection_rate) == type(None):
                opt.injection_rate = self.parameters['injection_rate']
        if type(opt.dealias_type) == type(None):
            opt.dealias_type = self.parameters['dealias_type']
        if (opt.nx > opt.n or
            opt.ny > opt.n or
            opt.nz > opt.n):
            opt.n = min(opt.nx, opt.ny, opt.nz)
            print("Warning: '-n' parameter changed to minimum of nx, ny, nz. This affects the computation of nu.")
        self.parameters['nu'] = (opt.kMeta * 2 / opt.n)**(4./3)
        # check value of kMax
        kM = opt.n * 0.5
        if opt.dealias_type == 1:
            kM *= 0.8
        self.set_host_info(TurTLE.host_info)
        if type(opt.environment) != type(None):
            self.host_info['environment'] = opt.environment
        self.pars_from_namespace(opt)
        return opt

    def write_par(
            self,
            iter0 = 0):
        assert (self.parameters['niter_todo'] % self.parameters['niter_stat'] == 0)
        assert (self.parameters['niter_todo'] % self.parameters['niter_out']  == 0)
        assert (self.parameters['niter_out']  % self.parameters['niter_stat'] == 0)
        _code.write_par(self, iter0 = iter0)
        with h5py.File(self.get_data_file_name(), 'r+') as ofile:
            ofile['code_info/exec_name'] = self.name
            kspace = self.get_kspace()
            for k in kspace.keys():
                ofile['kspace/' + k] = kspace[k]
            nshells = kspace['nshell'].shape[0]
            kspace = self.get_kspace()
            nshells = kspace['nshell'].shape[0]
            scal_stat_datasets = ['scalar']
            for k in scal_stat_datasets:
                time_chunk = 2**20//(8*nshells)
                time_chunk = max(time_chunk, 1)
                ofile.create_dataset('statistics/spectra/' + k + '_' + k,
                                     (1, nshells),
                                     chunks = (time_chunk, nshells),
                                     maxshape = (None, nshells),
                                     dtype = np.float64)
                time_chunk = 2**20//(8*10)
                time_chunk = max(time_chunk, 1)
                a = ofile.create_dataset('statistics/moments/' + k,
                                     (1, 10),
                                     chunks = (time_chunk, 10),
                                     maxshape = (None, 10),
                                     dtype = np.float64)
                time_chunk = 2**20//(8*self.parameters['histogram_bins'])
                time_chunk = max(time_chunk, 1)
                ofile.create_dataset('statistics/histograms/' + k,
                                     (1,
                                      self.parameters['histogram_bins']),
                                     chunks = (time_chunk,
                                               self.parameters['histogram_bins']),
                                     maxshape = (None,
                                                 self.parameters['histogram_bins']),
                                     dtype = np.int64)
            ofile['checkpoint'] = int(0)
        return None

    def launch(
            self,
            args = [],
            **kwargs):
        self.check_current_vorticity_exists = False
        opt = self.prepare_launch(args = args)
        if not os.path.exists(self.get_data_file_name()):
            self.write_par()
        self.run(
                nb_processes = opt.nb_processes,
                nb_threads_per_process = opt.nb_threads_per_process,
                njobs = opt.njobs,
                hours = opt.minutes // 60,
                minutes = opt.minutes % 60,
                no_submit = opt.no_submit,
                no_debug = not opt.debug)
        return None

def main():
    bla = ADNS('Kraichnan_scalar_v1')
    bla.launch(sys.argv[1:])
    return None

if __name__ == '__main__':
    main()
