import os
import sys

import numpy as np
import h5py

import TurTLE
import TurTLE.DNS

cpp_location = os.path.join(
        TurTLE.data_dir, 'particle_set')


class ADNS(TurTLE.DNS):
    def __init__(
            self,
            **kwargs):
        TurTLE.DNS.__init__(self, **kwargs)
        self.list_of_particle_codes.append('NSVEparticle_set')
        self.extra_parameters['NSVEparticle_set'] = {}
        return None
    def write_src(
            self):
        self.version_message = (
                '/***********************************************************************\n' +
                '* this code automatically generated by TurTLE\n' +
                '* version {0}\n'.format(TurTLE.__version__) +
                '***********************************************************************/\n\n\n')
        self.include_list = [
                '"base.hpp"',
                '"scope_timer.hpp"',
                '"fftw_interface.hpp"',
                '"full_code/main_code.hpp"',
                '"full_code/NSVEparticles.hpp"',
                '<cmath>',
                '<iostream>',
                '<hdf5.h>',
                '<string>',
                '<cstring>',
                '<fftw3-mpi.h>',
                '<omp.h>',
                '<cfenv>',
                '<cstdlib>']
        self.main = """
            int main(int argc, char *argv[])
            {{
                bool fpe = (
                    (getenv("BFPS_FPE_OFF") == nullptr) ||
                    (getenv("BFPS_FPE_OFF") != std::string("TRUE")));
                return main_code< {0} >(argc, argv, fpe);
            }}
            """.format(self.dns_type + '<{0}>'.format(self.C_field_dtype))
        self.includes = '\n'.join(
                ['#include ' + hh
                 for hh in self.include_list])
        self.name = 'NSVEparticle_set'
        self.dns_type = 'NSVEparticle_set'
        self.definitions = open(
            os.path.join(
                cpp_location, 'NSVEparticle_set.hpp'), 'r').read()
        self.definitions += open(
            os.path.join(
                cpp_location, 'NSVEparticle_set.cpp'), 'r').read()
        with open(self.name + '.cpp', 'w') as outfile:
            outfile.write(self.version_message + '\n\n')
            outfile.write(self.includes + '\n\n')
            outfile.write(self.definitions + '\n\n')
            outfile.write(self.main + '\n')
        self.check_current_vorticity_exists = True
        return None

    def write_par(
            self,
            iter0 = 0):
        TurTLE.DNS.write_par(self, iter0 = iter0)
        with h5py.File(self.get_data_file_name(), 'r+') as ofile:
            #TODO move collision stats to particle stats
            for i in range(self.parameters['niter_out']):
                ofile.create_dataset('statistics/collisions/'+str(i),
                                     shape=(1,),
                                     dtype = np.int64)
        return None

    def launch(
            self,
            args = [],
            **kwargs):
        opt = self.prepare_launch(args = args)
        if not os.path.exists(self.get_data_file_name()):
            self.write_par()
            self.generate_initial_condition(opt = opt)
            with h5py.File(self.get_particle_file_name(), 'w') as particle_file:
                particle_file.create_group('tracers0/position')
                particle_file.create_group('tracers0/velocity')
                particle_file.create_group('collisions0')
            self.generate_tracer_state(integration_steps = self.parameters['tracers0_integration_steps'],
                                       rseed = opt.particle_rand_seed,
                                       ncomponents = 3)
        self.run(
                nb_processes = opt.nb_processes,
                nb_threads_per_process = opt.nb_threads_per_process,
                njobs = opt.njobs,
                hours = opt.minutes // 60,
                minutes = opt.minutes % 60,
                no_submit = opt.no_submit,
                no_debug = not opt.debug)
        return None

def main():
    bla = ADNS()
    bla.launch([
        'NSVEparticle_set',
        '--niter_todo', '200',
        '--niter_stat', '20',
        '--niter_out', '200',
        '--dt', '0.0001',
        '--nparticles', '1000',
        '--tracers0_integration_steps', '0',
        '--tracers0_neighbours', '3',
        '--tracers0_smoothness', '2',
        '--tracers0_cutoff', '0.920272'
        ] + sys.argv[1:])
    assert(bla.parameters['niter_part'] == 1)
    df = bla.get_particle_file()
    x = []
    v = []
    for ii in range(bla.parameters['niter_out']):
        x.append(df['tracers0/position/{0}'.format(ii)][()])
        v.append(df['tracers0/velocity/{0}'.format(ii)][()])
    df.close()
    x = np.array(x)
    v = np.array(v)
    d1x = (x[2:] - x[:-2]) / (2*bla.parameters['dt'])
    vv = v[1:-1]
    diff = vv - d1x
    relative_error = (np.sum(diff**2, axis = 2) / np.sum(vv**2, axis = 2))**0.5
    tindex, trajindex = np.unravel_index(
            np.argmax(relative_error),
            relative_error.shape)
    try:
        import matplotlib.pyplot as plt
        f = plt.figure()
        a = f.add_subplot(211)
        a.plot(vv[:, trajindex])
        a.plot(d1x[:, trajindex], dashes = (2, 2))
        a = f.add_subplot(212)
        a.plot(x[:, trajindex])
        a.set_title('maximum error for (tindex, trajindex) = ({0}, {1})'.format(tindex, trajindex))
        f.tight_layout()
        f.savefig('traj.pdf')
        plt.close(f)
    except ImportError:
        print('couldn\'t import matplotlib, no figure')
    print(np.max(relative_error))
    assert(np.max(relative_error) < 1e-4)
    return None

if __name__ == '__main__':
    main()
