import os
import sys

import numpy as np
import h5py

import TurTLE
import TurTLE.DNS

cpp_location = os.path.join(
        TurTLE.data_dir, 'particle_set')

class ADNS(TurTLE.DNS):
    def __init__(
            self,
            **kwargs):
        TurTLE.DNS.__init__(self, **kwargs)
        self.list_of_particle_codes.append('particle_deleter')
        self.extra_parameters['particle_deleter'] = {}
        return None
    def custom_parameter_update(self):
        self.parameters['niter_out'] = int(1)
        self.extra_parameters['particle_deleter']['nparticles'] = int(100)
        self.extra_parameters['particle_deleter']['tracers0_integration_steps'] = int(0)
        self.extra_parameters['particle_deleter']['tracers0_neighbours'] = int(3)
        self.extra_parameters['particle_deleter']['tracers0_smoothness'] = int(2)
        return None
    def write_src(
            self):
        self.version_message = (
                '/***********************************************************************\n' +
                '* this code automatically generated by TurTLE\n' +
                '* version {0}\n'.format(TurTLE.__version__) +
                '***********************************************************************/\n\n\n')
        self.include_list = [
                '"base.hpp"',
                '"scope_timer.hpp"',
                '"fftw_interface.hpp"',
                '"full_code/main_code.hpp"',
                '"full_code/NSVEparticles.hpp"',
                '<cmath>',
                '<iostream>',
                '<hdf5.h>',
                '<string>',
                '<cstring>',
                '<fftw3-mpi.h>',
                '<omp.h>',
                '<cfenv>',
                '<cstdlib>']
        self.main = """
            int main(int argc, char *argv[])
            {{
                bool fpe = (
                    (getenv("BFPS_FPE_OFF") == nullptr) ||
                    (getenv("BFPS_FPE_OFF") != std::string("TRUE")));
                return main_code< {0} >(argc, argv, fpe);
            }}
            """.format(self.dns_type + '<{0}>'.format(self.C_field_dtype))
        self.includes = '\n'.join(
                ['#include ' + hh
                 for hh in self.include_list])
        self.name = 'particle_deleter'
        self.dns_type = 'particle_deleter'
        self.definitions = open(
            os.path.join(
                cpp_location, self.dns_type + '.hpp'), 'r').read()
        self.definitions += open(
            os.path.join(
                cpp_location, self.dns_type + '.cpp'), 'r').read()
        with open(self.name + '.cpp', 'w') as outfile:
            outfile.write(self.version_message + '\n\n')
            outfile.write(self.includes + '\n\n')
            outfile.write(self.definitions + '\n\n')
            outfile.write(self.main + '\n')
        self.check_current_vorticity_exists = True
        return None

    def write_par(
            self,
            iter0 = 0):
        TurTLE.DNS.write_par(self, iter0 = iter0)
        with h5py.File(self.get_data_file_name(), 'r+') as ofile:
            #TODO move collision stats to particle stats
            for i in range(self.parameters['niter_out']):
                ofile.create_dataset('statistics/collisions/'+str(i),
                                     shape=(1,),
                                     dtype = np.int64)
        return None

    def launch(
            self,
            args = [],
            **kwargs):
        opt = self.prepare_launch(args = args)
        if not os.path.exists(self.get_data_file_name()):
            self.write_par()
            self.generate_initial_condition(opt = opt)
            with h5py.File(self.get_particle_file_name(), 'w') as particle_file:
                particle_file.create_group('tracers0/position')
                particle_file.create_group('tracers0/velocity')
                particle_file.create_group('collisions0')
            with h5py.File('particle_index_before.h5', 'w') as particle_file:
                particle_file.create_group('tracers0/index')
            with h5py.File('particle_index_after.h5', 'w') as particle_file:
                particle_file.create_group('tracers0/index')
            with h5py.File('particle_subset_index.h5', 'w') as particle_file:
                particle_file.create_group('tracers_subset/index')
            self.generate_tracer_state(integration_steps = self.parameters['tracers0_integration_steps'],
                                       rseed = opt.particle_rand_seed,
                                       ncomponents = 3)
        self.run(
                nb_processes = opt.nb_processes,
                nb_threads_per_process = opt.nb_threads_per_process,
                njobs = opt.njobs,
                hours = opt.minutes // 60,
                minutes = opt.minutes % 60,
                no_submit = opt.no_submit,
                no_debug = not opt.debug)
        return None

def main():
    #####################################################################
    ## run c++ test
    bla = ADNS()
    bla.launch(sys.argv[1:])
    #####################################################################

    #####################################################################
    ## python routine replicating deleting pattern of the test in c++
    def delete_index(before_py, iteration):
        index2 =  (iteration*3)%len(before_py)
        index_to_delete = np.array([1, index2])
        if index2==1:
            index_to_delete = np.array([1, 3])
        after_py = np.delete(before_py, index_to_delete)
        return before_py, after_py
    #####################################################################

    #####################################################################
    ## perform sanity check
    f = h5py.File(bla.simname + '.h5', 'r')
    p = f['parameters']
    #initialising list of indeces
    before_py = np.arange(0,np.array(p['nparticles']))
    #comparing list of particles for each iteration
    for i in range(np.array(p['niter_todo'])):
        #apply python routine
        before_py, after_py = delete_index(before_py, i)
        #load turtle indeces
        before_turtle = h5py.File('particle_index_before.h5','r')
        after_turtle = h5py.File('particle_index_after.h5','r')
        subset_turtle = h5py.File('particle_subset_index.h5', 'r')
        index_before_turtle = np.array(before_turtle['tracers0/index/'+str(i)])
        index_after_turtle = np.array(after_turtle['tracers0/index/'+str(i)])
        index_subset = np.array(subset_turtle['tracers_subset/index/'+str(i)])
        index_before_turtle = index_before_turtle.flatten()
        index_after_turtle = index_after_turtle.flatten()
        index_subset = index_subset.flatten()
        print('testing labels iteration {0}'.format(i))
        assert(np.all(before_py==index_before_turtle) and
               np.all(after_py==index_after_turtle))
        print('testing subset labels iteration {0}'.format(i))
        assert(np.all(after_py[[2, 4]]==index_subset))
        if(len(after_py)==2):
            break
        before_py = after_py
    #####################################################################
    return None

if __name__ == '__main__':
    main()
