/**********************************************************************
*                                                                     *
*  Copyright 2015 the TurTLE team                                     *
*                                                                     *
*  This file is part of TurTLE.                                       *
*                                                                     *
*  TurTLE is free software: you can redistribute it and/or modify     *
*  it under the terms of the GNU General Public License as published  *
*  by the Free Software Foundation, either version 3 of the License,  *
*  or (at your option) any later version.                             *
*                                                                     *
*  TurTLE is distributed in the hope that it will be useful,          *
*  but WITHOUT ANY WARRANTY; without even the implied warranty of     *
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      *
*  GNU General Public License for more details.                       *
*                                                                     *
*  You should have received a copy of the GNU General Public License  *
*  along with TurTLE.  If not, see <http://www.gnu.org/licenses/>     *
*                                                                     *
* Contact: Cristian.Lalescu@ds.mpg.de                                 *
*                                                                     *
**********************************************************************/



#include "fftw_tools.hpp"
#include "vorticity_equation.hpp"
#include "scope_timer.hpp"
#include "shared_array.hpp"

#include <sys/stat.h>

template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::impose_zero_modes()
{
    TIMEZONE("vorticity_equation::impose_zero_modes");
    this->u->impose_zero_mode();
    this->v[0]->impose_zero_mode();
    this->v[1]->impose_zero_mode();
    this->v[2]->impose_zero_mode();
}

template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::update_checkpoint()
{
    std::string fname = this->get_current_fname();
    if (this->kk->layout->myrank == 0)
    {
        bool file_exists = false;
        {
            struct stat file_buffer;
            file_exists = (stat(fname.c_str(), &file_buffer) == 0);
        }
        if (file_exists)
        {
            // check how many fields there are in the checkpoint file
            // increment checkpoint if needed
            hsize_t fields_stored;
            hid_t fid, group_id;
            fid = H5Fopen(fname.c_str(), H5F_ACC_RDONLY, H5P_DEFAULT);
            bool vorticity_group_exists = H5Lexists(
                    fid,
                    "vorticity",
                    H5P_DEFAULT);
            if (vorticity_group_exists) {
                bool complex_vorticity_group_exists = H5Lexists(
                    fid,
                    "vorticity/complex",
                    H5P_DEFAULT);
                if (complex_vorticity_group_exists) {
                    group_id = H5Gopen(fid, "vorticity/complex", H5P_DEFAULT);
                    H5Gget_num_objs(
                            group_id,
                            &fields_stored);
                    bool dset_exists = H5Lexists(
                            group_id,
                            std::to_string(this->iteration).c_str(),
                            H5P_DEFAULT);
                    H5Gclose(group_id);
                    if ((int(fields_stored) >= this->checkpoints_per_file) &&
                        !dset_exists)
                        this->checkpoint++;
                }
            }
            H5Fclose(fid);
        }
        else
        {
            // create file, create fields_stored dset
            hid_t fid = H5Fcreate(
                    fname.c_str(),
                    H5F_ACC_EXCL,
                    H5P_DEFAULT,
                    H5P_DEFAULT);
            hid_t gg = H5Gcreate(
                    fid,
                    "vorticity",
                    H5P_DEFAULT,
                    H5P_DEFAULT,
                    H5P_DEFAULT);
            hid_t ggg = H5Gcreate(
                    gg,
                    "complex",
                    H5P_DEFAULT,
                    H5P_DEFAULT,
                    H5P_DEFAULT);
            H5Gclose(ggg);
            H5Gclose(gg);
            H5Fclose(fid);
        }
    }
    MPI_Bcast(&this->checkpoint, 1, MPI_INT, 0, this->kk->layout->comm);
}

template <class rnumber,
          field_backend be>
vorticity_equation<rnumber, be>::vorticity_equation(
        const char *NAME,
        const int nx,
        const int ny,
        const int nz,
        const MPI_Comm comm,
        double DKX,
        double DKY,
        double DKZ,
        unsigned FFTW_PLAN_RIGOR)
{
    TIMEZONE("vorticity_equation::vorticity_equation");
    /* initialize name and basic stuff */
    strncpy(this->name, NAME, 256);
    this->name[255] = '\0';
    this->iteration = 0;
    this->checkpoint = 0;

    /* initialize fields */
    this->cvorticity = new field<rnumber, be, THREE>(
            nx, ny, nz, comm, FFTW_PLAN_RIGOR);
    this->rvorticity = new field<rnumber, be, THREE>(
            nx, ny, nz, comm, FFTW_PLAN_RIGOR);
    this->v[1] = new field<rnumber, be, THREE>(
            nx, ny, nz, comm, FFTW_PLAN_RIGOR);
    this->v[2] = new field<rnumber, be, THREE>(
            nx, ny, nz, comm, FFTW_PLAN_RIGOR);
    this->v[0] = this->cvorticity;
    this->v[3] = this->cvorticity;

    this->cvelocity = new field<rnumber, be, THREE>(
            nx, ny, nz, comm, FFTW_PLAN_RIGOR);
    this->u = this->cvelocity;

    /* initialize kspace */
    this->kk = new kspace<be, SMOOTH>(
            this->cvorticity->clayout, DKX, DKY, DKZ);

    /* ``physical'' parameters etc, initialized here just in case */

    this->nu = 0.1;
    this->fmode = 1;
    this->famplitude = 1.0;
    this->friction_coefficient = 1.0;
    this->fk0  = 2.0;
    this->fk1 = 4.0;
    this->dt = 0.1;
}

template <class rnumber,
          field_backend be>
vorticity_equation<rnumber, be>::~vorticity_equation() noexcept(false)
{
    TIMEZONE("vorticity_equation::~vorticity_equation");
    delete this->kk;
    delete this->cvorticity;
    delete this->rvorticity;
    delete this->v[1];
    delete this->v[2];
    delete this->cvelocity;
}

template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::compute_vorticity()
{
    TIMEZONE("vorticity_equation::compute_vorticity");
    this->cvorticity->real_space_representation = false;
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const ptrdiff_t xindex,
                    const ptrdiff_t yindex,
                    const ptrdiff_t zindex,
                    const double k2){
        if (k2 <= this->kk->kM2)
        {
            this->cvorticity->cval(cindex,0,0) = -(this->kk->ky[yindex]*this->u->cval(cindex,2,1) - this->kk->kz[zindex]*this->u->cval(cindex,1,1));
            this->cvorticity->cval(cindex,0,1) =  (this->kk->ky[yindex]*this->u->cval(cindex,2,0) - this->kk->kz[zindex]*this->u->cval(cindex,1,0));
            this->cvorticity->cval(cindex,1,0) = -(this->kk->kz[zindex]*this->u->cval(cindex,0,1) - this->kk->kx[xindex]*this->u->cval(cindex,2,1));
            this->cvorticity->cval(cindex,1,1) =  (this->kk->kz[zindex]*this->u->cval(cindex,0,0) - this->kk->kx[xindex]*this->u->cval(cindex,2,0));
            this->cvorticity->cval(cindex,2,0) = -(this->kk->kx[xindex]*this->u->cval(cindex,1,1) - this->kk->ky[yindex]*this->u->cval(cindex,0,1));
            this->cvorticity->cval(cindex,2,1) =  (this->kk->kx[xindex]*this->u->cval(cindex,1,0) - this->kk->ky[yindex]*this->u->cval(cindex,0,0));
        }
        else
            std::fill_n((rnumber*)(this->cvorticity->get_cdata()+3*cindex), 6, 0.0);
    }
    );
    this->cvorticity->symmetrize();
}

template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::compute_velocity(field<rnumber, be, THREE> *vorticity)
{
    TIMEZONE("vorticity_equation::compute_velocity");
    *(this->u) = 0.0;
    this->u->real_space_representation = false;
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const ptrdiff_t xindex,
                    const ptrdiff_t yindex,
                    const ptrdiff_t zindex,
                    const double k2){
        const double kk2 = (k2 > 0) ? k2 : 1.;
        if (k2 <= this->kk->kM2)
        {
            this->u->cval(cindex,0,0) = -(this->kk->ky[yindex]*vorticity->cval(cindex,2,1) - this->kk->kz[zindex]*vorticity->cval(cindex,1,1)) / kk2;
            this->u->cval(cindex,0,1) =  (this->kk->ky[yindex]*vorticity->cval(cindex,2,0) - this->kk->kz[zindex]*vorticity->cval(cindex,1,0)) / kk2;
            this->u->cval(cindex,1,0) = -(this->kk->kz[zindex]*vorticity->cval(cindex,0,1) - this->kk->kx[xindex]*vorticity->cval(cindex,2,1)) / kk2;
            this->u->cval(cindex,1,1) =  (this->kk->kz[zindex]*vorticity->cval(cindex,0,0) - this->kk->kx[xindex]*vorticity->cval(cindex,2,0)) / kk2;
            this->u->cval(cindex,2,0) = -(this->kk->kx[xindex]*vorticity->cval(cindex,1,1) - this->kk->ky[yindex]*vorticity->cval(cindex,0,1)) / kk2;
            this->u->cval(cindex,2,1) =  (this->kk->kx[xindex]*vorticity->cval(cindex,1,0) - this->kk->ky[yindex]*vorticity->cval(cindex,0,0)) / kk2;
        }
        else
            std::fill_n((rnumber*)(this->u->get_cdata()+3*cindex), 6, 0.0);
    }
    );
    if (this->kk->layout->myrank == 0)
        std::fill_n((rnumber*)this->u->get_cdata(), 6, 0.0);
    this->u->symmetrize();
}

template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::add_Kolmogorov_forcing(
        field<rnumber, be, THREE> *dst,
        const int fmode,
        const double famplitude)
{
    TIMEZONE("vorticity_equation::add_Kolmogorov_forcing");
    ptrdiff_t cindex;
    if (dst->clayout->myrank == dst->clayout->rank[0][fmode])
    {
        cindex = dst->get_cindex(0, (fmode - dst->clayout->starts[0]), 0);
        dst->cval(cindex,2, 0) -= famplitude/2;
    }
    if (dst->clayout->myrank == dst->clayout->rank[0][dst->clayout->sizes[0] - fmode])
    {
        cindex = dst->get_cindex(0, (dst->clayout->sizes[0] - fmode - dst->clayout->starts[0]), 0);
        dst->cval(cindex, 2, 0) -= famplitude/2;
    }
}

template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::add_field_band(
        field<rnumber, be, THREE> *dst,
        field<rnumber, be, THREE> *src,
        const double k0, const double k1,
        const double prefactor)
{
    TIMEZONE("vorticity_equation::add_field_band");
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const ptrdiff_t xindex,
                    const ptrdiff_t yindex,
                    const ptrdiff_t zindex){
        const double knorm = std::sqrt(
                this->kk->kx[xindex]*this->kk->kx[xindex] +
                this->kk->ky[yindex]*this->kk->ky[yindex] +
                this->kk->kz[zindex]*this->kk->kz[zindex]);
        if ((k0 <= knorm) &&
            (k1 >= knorm))
            for (int c=0; c<3; c++)
                for (int i=0; i<2; i++)
                    dst->cval(cindex,c,i) += prefactor*src->cval(cindex,c,i);
    }
    );
}

template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::add_forcing(
        field<rnumber, be, THREE> *dst,
        field<rnumber, be, THREE> *vort_field,
        double t)
{
    TIMEZONE("vorticity_equation::add_forcing");
    if (this->forcing_type == std::string("Kolmogorov"))
    {
        this->add_Kolmogorov_forcing(dst, this->fmode, this->famplitude);
        return;
    }
    if (this->forcing_type == std::string("2Kolmogorov"))
    {
        // 2 Kolmogorov forces
        // first one wavenumber fk0, amplitude 1 - A
        double amplitude = 1 - this->famplitude;
        int fmode = int(this->fk0 / this->kk->dky);
        this->add_Kolmogorov_forcing(dst, fmode, amplitude);
        // second one wavenumber fk1, amplitude A
        amplitude = this->famplitude * pow(int(this->fk1) / double(int(this->fk0)), 3);
        fmode = int(this->fk1 / this->kk->dky);
        this->add_Kolmogorov_forcing(dst, fmode, amplitude);
        return;
    }
    if (this->forcing_type == std::string("Kolmogorov_and_drag"))
    {
        this->add_Kolmogorov_forcing(dst, this->fmode, this->famplitude);
        this->add_field_band(
                dst, vort_field,
                this->fk0, this->fk1,
                -this->friction_coefficient);
        return;
    }
    if (this->forcing_type == std::string("Kolmogorov_and_compensated_drag"))
    {
        double amplitude = this->famplitude * (
                1 + this->friction_coefficient / std::sqrt(this->fmode  * this->famplitude));
        this->add_Kolmogorov_forcing(dst, this->fmode, amplitude);
        this->add_field_band(
                dst, vort_field,
                this->fk0, this->fk1,
                -this->friction_coefficient);
        return;
    }
    if (this->forcing_type == std::string("linear"))
    {
        this->add_field_band(
                dst, vort_field,
                this->fk0, this->fk1,
                this->famplitude);
        return;
    }
    if (this->forcing_type == std::string("fixed_energy_injection_rate") ||
        this->forcing_type == std::string("fixed_energy_injection_rate_and_drag") ||
        this->forcing_type == std::string("sinusoidal_energy_injection_rate"))
    {
        // first, compute energy in shell
        shared_array<double> local_energy_in_shell(
                1,
                shared_array_zero_initializer<double, 1>);
        double energy_in_shell = 0;
        this->kk->CLOOP(
                    [&](const ptrdiff_t cindex,
                        const ptrdiff_t xindex,
                        const ptrdiff_t yindex,
                        const ptrdiff_t zindex,
                        const double k2,
                        const int nxmodes) {
            const double knorm = std::sqrt(k2);
            if ((k2 > 0) &&
                (this->fk0 <= knorm) &&
                (this->fk1 >= knorm)) {
                    *local_energy_in_shell.getMine() += nxmodes*(
                            vort_field->cval(cindex, 0, 0)*vort_field->cval(cindex, 0, 0) + vort_field->cval(cindex, 0, 1)*vort_field->cval(cindex, 0, 1) +
                            vort_field->cval(cindex, 1, 0)*vort_field->cval(cindex, 1, 0) + vort_field->cval(cindex, 1, 1)*vort_field->cval(cindex, 1, 1) +
                            vort_field->cval(cindex, 2, 0)*vort_field->cval(cindex, 2, 0) + vort_field->cval(cindex, 2, 1)*vort_field->cval(cindex, 2, 1)
                            ) / k2;
                    }
        });
        local_energy_in_shell.mergeParallel();
        MPI_Allreduce(
                local_energy_in_shell.getMasterData(),
                &energy_in_shell,
                1,
                MPI_DOUBLE,
                MPI_SUM,
                vort_field->comm);
        // we should divide by 2, if we wanted energy;
        // but then we would need to multiply the amplitude by 2 anyway,
        // because what we really care about is force dotted into velocity,
        // without the division by 2.

        // now, modify amplitudes
        if (energy_in_shell < 10*std::numeric_limits<rnumber>::epsilon())
            energy_in_shell = 1;
        double temp_famplitude = 0;
        if (this->forcing_type == std::string("sinusoidal_energy_injection_rate")){
            temp_famplitude = (this->injection_rate + this->variation_strength
                                *std::sin(t/this->variation_time_scale))
                                / energy_in_shell;
        }else{
            temp_famplitude = this->injection_rate / energy_in_shell;
        }
        this->add_field_band(
                dst, vort_field,
                this->fk0, this->fk1,
                temp_famplitude);
        // and add drag if desired
        if (this->forcing_type == std::string("fixed_energy_injection_rate_and_drag"))
            this->add_field_band(
                    dst, vort_field,
                    this->fmode, this->fmode + (this->fk1 - this->fk0),
                    -this->friction_coefficient);
        return;
    }
    if (this->forcing_type == std::string("fixed_energy"))
        return;
    else
    {
        DEBUG_MSG("unknown forcing type printed on next line\n%s", this->forcing_type.c_str());
        throw std::invalid_argument("unknown forcing type");
    }
}

template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::impose_forcing(
        field<rnumber, be, THREE> *onew,
        field<rnumber, be, THREE> *oold)
{
    TIMEZONE("vorticity_equation::impose_forcing");
    if (this->forcing_type == std::string("fixed_energy"))
    {
        // first, compute energy in shell
        shared_array<double> local_energy_in_shell(
                1,
                shared_array_zero_initializer<double, 1>);
        shared_array<double> local_total_energy(
                1,
                shared_array_zero_initializer<double, 1>);
        double energy_in_shell, total_energy;
        this->kk->CLOOP(
                    [&](const ptrdiff_t cindex,
                        const ptrdiff_t xindex,
                        const ptrdiff_t yindex,
                        const ptrdiff_t zindex,
                        const double k2,
                        const int nxmodes){
            if (k2 > 0)
            {
                const double mode_energy = nxmodes*(
                            onew->cval(cindex, 0, 0)*onew->cval(cindex, 0, 0) + onew->cval(cindex, 0, 1)*onew->cval(cindex, 0, 1) +
                            onew->cval(cindex, 1, 0)*onew->cval(cindex, 1, 0) + onew->cval(cindex, 1, 1)*onew->cval(cindex, 1, 1) +
                            onew->cval(cindex, 2, 0)*onew->cval(cindex, 2, 0) + onew->cval(cindex, 2, 1)*onew->cval(cindex, 2, 1)
                            ) / k2;
                *local_total_energy.getMine() += mode_energy;
                double knorm = std::sqrt(k2);
                if ((this->fk0 <= knorm) && (this->fk1 >= knorm))
                    *local_energy_in_shell.getMine() += mode_energy;
            }
        });
        local_total_energy.mergeParallel();
        local_energy_in_shell.mergeParallel();
        MPI_Allreduce(
                local_energy_in_shell.getMasterData(),
                &energy_in_shell,
                1,
                MPI_DOUBLE,
                MPI_SUM,
                onew->comm);
        MPI_Allreduce(
                local_total_energy.getMasterData(),
                &total_energy,
                1,
                MPI_DOUBLE,
                MPI_SUM,
                onew->comm);
        // divide by 2, because we want energy
        total_energy /= 2;
        energy_in_shell /= 2;
        // now, add forcing term
        // see Michael's thesis, page 38
        double temp_famplitude = std::sqrt(
                (this->energy - total_energy + energy_in_shell)
               / energy_in_shell);
        this->kk->CLOOP(
                    [&](const ptrdiff_t cindex,
                        const ptrdiff_t xindex,
                        const ptrdiff_t yindex,
                        const ptrdiff_t zindex,
                        const double k2){
            const double knorm = std::sqrt(k2);
            if ((this->fk0 <= knorm) &&
                (this->fk1 >= knorm))
                for (int c=0; c<3; c++)
                    for (int i=0; i<2; i++)
                        onew->cval(cindex,c,i) *= temp_famplitude;
        }
        );
        return;
    }
}

template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::omega_nonlin(
        int src, double t)
{
    //DEBUG_MSG("vorticity_equation::omega_nonlin(%d)\n", src);
    assert(src >= 0 && src < 3);
    this->compute_velocity(this->v[src]);
    /* get fields from Fourier space to real space */
    this->u->ift();
    this->rvorticity->real_space_representation = false;
    *this->rvorticity = this->v[src]->get_cdata();
    this->rvorticity->ift();
    /* compute cross product $u \times \omega$, and normalize */
    this->u->RLOOP(
                [&](const ptrdiff_t rindex,
                    const ptrdiff_t xindex,
                    const ptrdiff_t yindex,
                    const ptrdiff_t zindex){
        rnumber tmp[3];
        for (int cc=0; cc<3; cc++)
            tmp[cc] = (this->u->rval(rindex,(cc+1)%3)*this->rvorticity->rval(rindex,(cc+2)%3) -
                       this->u->rval(rindex,(cc+2)%3)*this->rvorticity->rval(rindex,(cc+1)%3));
        for (int cc=0; cc<3; cc++)
            this->u->rval(rindex,cc) = tmp[cc] / this->u->npoints;
    }
    );
    /* go back to Fourier space */
    //this->clean_up_real_space(this->ru, 3);
    this->u->dft();
    this->kk->template dealias<rnumber, THREE>(this->u->get_cdata());
    /* $\imath k \times Fourier(u \times \omega)$ */
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const ptrdiff_t xindex,
                    const ptrdiff_t yindex,
                    const ptrdiff_t zindex){
        // this tmp needs to be defined as const
        const rnumber tmp[3][2] = {
                {(rnumber)(-(this->kk->ky[yindex]*this->u->cval(cindex,2,1) - this->kk->kz[zindex]*this->u->cval(cindex,1,1))),
                 (rnumber)( (this->kk->ky[yindex]*this->u->cval(cindex,2,0) - this->kk->kz[zindex]*this->u->cval(cindex,1,0)))},
                {(rnumber)(-(this->kk->kz[zindex]*this->u->cval(cindex,0,1) - this->kk->kx[xindex]*this->u->cval(cindex,2,1))),
                 (rnumber)( (this->kk->kz[zindex]*this->u->cval(cindex,0,0) - this->kk->kx[xindex]*this->u->cval(cindex,2,0)))},
                {(rnumber)(-(this->kk->kx[xindex]*this->u->cval(cindex,1,1) - this->kk->ky[yindex]*this->u->cval(cindex,0,1))),
                 (rnumber)( (this->kk->kx[xindex]*this->u->cval(cindex,1,0) - this->kk->ky[yindex]*this->u->cval(cindex,0,0)))}
                };
        //rnumber tmp[3][2];
        //{
        //    tmp[0][0] = -(this->kk->ky[yindex]*this->u->cval(cindex,2,1) - this->kk->kz[zindex]*this->u->cval(cindex,1,1));
        //    tmp[1][0] = -(this->kk->kz[zindex]*this->u->cval(cindex,0,1) - this->kk->kx[xindex]*this->u->cval(cindex,2,1));
        //    tmp[2][0] = -(this->kk->kx[xindex]*this->u->cval(cindex,1,1) - this->kk->ky[yindex]*this->u->cval(cindex,0,1));
        //    tmp[0][1] =  (this->kk->ky[yindex]*this->u->cval(cindex,2,0) - this->kk->kz[zindex]*this->u->cval(cindex,1,0));
        //    tmp[1][1] =  (this->kk->kz[zindex]*this->u->cval(cindex,0,0) - this->kk->kx[xindex]*this->u->cval(cindex,2,0));
        //    tmp[2][1] =  (this->kk->kx[xindex]*this->u->cval(cindex,1,0) - this->kk->ky[yindex]*this->u->cval(cindex,0,0));
        //}
        //#pragma omp simd
        for (int cc=0; cc<3; cc++) for (int i=0; i<2; i++)
            this->u->cval(cindex, cc, i) = tmp[cc][i];
    }
    );
    this->kk->template force_divfree<rnumber>(this->u->get_cdata());
    this->u->symmetrize();
    this->add_forcing(this->u, this->v[src], t);
}

template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::Euler_step(double dt)
{
    TIMEZONE("vorticity_equation::step_Euler");
    *this->v[1] = 0.0;
    this->omega_nonlin(0, this->iteration*dt);
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const ptrdiff_t xindex,
                    const ptrdiff_t yindex,
                    const ptrdiff_t zindex,
                    const double k2){
        if (k2 <= this->kk->kM2)
        {
            const double factor0 = exp(-this->nu * k2 * dt);
            for (int cc=0; cc<3; cc++) for (int i=0; i<2; i++)
                this->v[1]->cval(cindex,cc,i) = (
                        this->v[0]->cval(cindex,cc,i) +
                        dt*this->u->cval(cindex,cc,i))*factor0;
        }
    }
    );
    this->impose_forcing(this->v[1], this->v[0]);

    *(this->cvorticity) = this->v[1]->get_cdata();

    this->kk->template force_divfree<rnumber>(this->cvorticity->get_cdata());
    this->cvorticity->symmetrize();
    this->iteration++;
}

template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::step(double dt)
{
    // This is an implementation of a memory-saving, third-order
    // Runge-Kutta method. See also
    // https://en.wikipedia.org/wiki/List_of_Runge%E2%80%93Kutta_methods#Explicit_methods
    // where it is called "Third-order Strong Stability Preserving Runge-Kutta (SSPRK3)".
    // The intermediate values of t passed to `omega_nonlin` can be read from
    // the corresponding Butcher tableau.
    TIMEZONE("vorticity_equation::step");
    *this->v[1] = 0.0;
    this->omega_nonlin(0, this->iteration*dt);
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const ptrdiff_t xindex,
                    const ptrdiff_t yindex,
                    const ptrdiff_t zindex,
                    const double k2){
        if (k2 <= this->kk->kM2)
        {
            const double factor0 = exp(-this->nu * k2 * dt);
            for (int cc=0; cc<3; cc++) for (int i=0; i<2; i++)
                this->v[1]->cval(cindex,cc,i) = (
                        this->v[0]->cval(cindex,cc,i) +
                        dt*this->u->cval(cindex,cc,i))*factor0;
        }
    }
    );
    this->impose_forcing(this->v[1], this->v[0]);
    this->kk->template force_divfree<rnumber>(this->v[1]->get_cdata());

    this->omega_nonlin(1, (this->iteration + 1)*dt);
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const ptrdiff_t xindex,
                    const ptrdiff_t yindex,
                    const ptrdiff_t zindex,
                    const double k2){
        if (k2 <= this->kk->kM2)
        {
            const double factor0 = exp(-this->nu * k2 * dt/2);
            const double factor1 = exp( this->nu * k2 * dt/2);
            for (int cc=0; cc<3; cc++) for (int i=0; i<2; i++)
                this->v[2]->cval(cindex, cc, i) = (
                        3*this->v[0]->cval(cindex,cc,i)*factor0 +
                        ( this->v[1]->cval(cindex,cc,i) +
                         dt*this->u->cval(cindex,cc,i))*factor1)*0.25;
        }
    }
    );
    this->impose_forcing(this->v[2], this->v[0]);
    this->kk->template force_divfree<rnumber>(this->v[2]->get_cdata());

    this->omega_nonlin(2, (this->iteration + 1./2.)*dt);
    // store old vorticity
    *this->v[1] = *this->v[0];
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const ptrdiff_t xindex,
                    const ptrdiff_t yindex,
                    const ptrdiff_t zindex,
                    const double k2){
        if (k2 <= this->kk->kM2)
        {
            const double factor0 = exp(-this->nu * k2 * dt * 0.5);
            for (int cc=0; cc<3; cc++) for (int i=0; i<2; i++)
                this->v[3]->cval(cindex,cc,i) = (
                        this->v[0]->cval(cindex,cc,i)*factor0 +
                        2*(this->v[2]->cval(cindex,cc,i) +
                           dt*this->u->cval(cindex,cc,i)))*factor0/3;
        }
    }
    );
    this->impose_forcing(this->v[0], this->v[1]);

    this->kk->template force_divfree<rnumber>(this->cvorticity->get_cdata());
    this->cvorticity->symmetrize();
    this->iteration++;
}

template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::compute_pressure(field<rnumber, be, ONE> *pressure)
{
    TIMEZONE("vorticity_equation::compute_pressure");
    *pressure = 0.0;
    /* assume velocity is already in real space representation */
    /* set the pressure representation to Fourier space */
    pressure->real_space_representation = false;

    this->v[1]->real_space_representation = true;
    /* diagonal terms 11 22 33 */
    *this->v[1] = 0.0;
    this->v[1]->RLOOP (
                [&](const ptrdiff_t rindex,
                    const hsize_t xindex,
                    const hsize_t yindex,
                    const hsize_t zindex){
        for (int cc=0; cc<3; cc++)
            this->v[1]->rval(rindex,cc) = this->u->rval(rindex,cc)*this->u->rval(rindex,cc);
        });
    this->v[1]->dft();
    //this->kk->template dealias<rnumber, THREE>(this->v[1]->get_cdata());
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const hsize_t xindex,
                    const hsize_t yindex,
                    const hsize_t zindex,
                    const double k2){
        if (k2 <= this->kk->kM2 && k2 > 0)
            for (int i=0; i<2; i++)
            {
                pressure->get_cdata()[cindex][i] =
                    -(this->kk->kx[xindex]*this->kk->kx[xindex]*this->v[1]->cval(cindex,0,i)
                    + this->kk->ky[yindex]*this->kk->ky[yindex]*this->v[1]->cval(cindex,1,i)
                    + this->kk->kz[zindex]*this->kk->kz[zindex]*this->v[1]->cval(cindex,2,i));
            }
        });
    /* off-diagonal terms 12 23 31 */
    this->v[1]->real_space_representation = true;
    *this->v[1] = 0.0;
    this->v[1]->RLOOP (
                [&](const ptrdiff_t rindex,
                    const hsize_t xindex,
                    const hsize_t yindex,
                    const hsize_t zindex){
        for (int cc=0; cc<3; cc++)
            this->v[1]->rval(rindex,cc) = this->u->rval(rindex,cc)*this->u->rval(rindex,(cc+1)%3);
        });
    this->v[1]->dft();
    //this->kk->template dealias<rnumber, THREE>(this->v[1]->get_cdata());
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const hsize_t xindex,
                    const hsize_t yindex,
                    const hsize_t zindex,
                    const double k2){
        if (k2 <= this->kk->kM2 && k2 > 0)
            for (int i=0; i<2; i++)
            {
                pressure->get_cdata()[cindex][i] -=
                    2*(this->kk->kx[xindex]*this->kk->ky[yindex]*this->v[1]->cval(cindex,0,i)
                     + this->kk->ky[yindex]*this->kk->kz[zindex]*this->v[1]->cval(cindex,1,i)
                     + this->kk->kz[zindex]*this->kk->kx[xindex]*this->v[1]->cval(cindex,2,i));
                pressure->get_cdata()[cindex][i] /= pressure->npoints*k2;
            }
        });
    this->kk->template dealias<rnumber, ONE>(pressure->get_cdata());
}


/** \brief Compute Lagrangian acceleration.
 *
 *  Acceleration is put in `acceleration` in the Fourier space representation.
 */
template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::compute_Lagrangian_acceleration(
        field<rnumber, be, THREE> *acceleration,
        field<rnumber, be, ONE> *pressure)
{
    bool own_pressure = false;
    if (pressure == NULL)
    {
        pressure = new field<rnumber, be, ONE>(
            this->cvelocity->rlayout->sizes[2],
            this->cvelocity->rlayout->sizes[1],
            this->cvelocity->rlayout->sizes[0],
            this->cvelocity->rlayout->comm,
            this->cvelocity->fftw_plan_rigor);
        own_pressure = true;
    }
    this->compute_velocity(this->cvorticity);
    this->cvelocity->ift();
    this->compute_pressure(pressure);
    this->compute_velocity(this->cvorticity);
    *acceleration = 0.0;
    acceleration->real_space_representation = false;
    // add curl of forcing
    *(this->v[1]) = 0.0;
    this->v[1]->real_space_representation = false;
    this->add_forcing(this->v[1], this->cvorticity, this->iteration*this->dt);
    // invert curl
    invert_curl(
        this->kk,
        this->v[1],
        acceleration);
    // loop over all modes, adding the rest of the terms
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const ptrdiff_t xindex,
                    const ptrdiff_t yindex,
                    const ptrdiff_t zindex,
                    const double k2){
        if (k2 <= this->kk->kM2)
        {
            // add dissipative term
            const double factor = this->nu*k2;
            for (int cc=0; cc<3; cc++)
                for (int i=0; i<2; i++)
                    acceleration->cval(cindex,cc, i) -= (
                        factor*this->u->cval(cindex,cc,i));

            // add nonlinear term
            acceleration->cval(cindex,0,0) += this->kk->kx[xindex]*pressure->cval(cindex,1);
            acceleration->cval(cindex,0,1) -= this->kk->kx[xindex]*pressure->cval(cindex,0);
            acceleration->cval(cindex,1,0) += this->kk->ky[yindex]*pressure->cval(cindex,1);
            acceleration->cval(cindex,1,1) -= this->kk->ky[yindex]*pressure->cval(cindex,0);
            acceleration->cval(cindex,2,0) += this->kk->kz[zindex]*pressure->cval(cindex,1);
            acceleration->cval(cindex,2,1) -= this->kk->kz[zindex]*pressure->cval(cindex,0);
        }
        });
    acceleration->symmetrize();
    if (own_pressure)
        delete pressure;
}

template <class rnumber,
          field_backend be>
void vorticity_equation<rnumber, be>::compute_Eulerian_acceleration(
        field<rnumber, be, THREE> *acceleration)
{
    this->compute_velocity(this->cvorticity);
    acceleration->real_space_representation = false;
    /* put in linear terms */
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const ptrdiff_t xindex,
                    const ptrdiff_t yindex,
                    const ptrdiff_t zindex,
                    const double k2){
        if (k2 <= this->kk->kM2)
        {
            ptrdiff_t tindex = 3*cindex;
            for (int cc=0; cc<3; cc++)
                for (int i=0; i<2; i++)
                    acceleration->get_cdata()[tindex+cc][i] = \
                        - this->nu*k2*this->cvelocity->get_cdata()[tindex+cc][i];
            if (this->forcing_type == std::string("linear"))
            {
                double knorm = std::sqrt(k2);
                if ((this->fk0 <= knorm) &&
                        (this->fk1 >= knorm))
                {
                    for (int c=0; c<3; c++)
                        for (int i=0; i<2; i++)
                            acceleration->get_cdata()[tindex+c][i] += \
                                this->famplitude*this->cvelocity->get_cdata()[tindex+c][i];
                }
            }
        }
    }
    );
    this->cvelocity->ift();
    /* compute uu */
    /* 11 22 33 */
    this->v[1]->real_space_representation = true;
    this->cvelocity->RLOOP (
                [&](ptrdiff_t rindex,
                    ptrdiff_t xindex,
                    ptrdiff_t yindex,
                    ptrdiff_t zindex){
        //ptrdiff_t tindex = 3*rindex;
        for (int cc=0; cc<3; cc++)
            this->v[1]->rval(rindex,cc) = \
                this->cvelocity->rval(rindex,cc)*this->cvelocity->rval(rindex,cc) / this->cvelocity->npoints;
    }
    );
    this->v[1]->dft();
    this->kk->template dealias<rnumber, THREE>(this->v[1]->get_cdata());
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const ptrdiff_t xindex,
                    const ptrdiff_t yindex,
                    const ptrdiff_t zindex,
                    const double k2){
        if (k2 <= this->kk->kM2)
        {
            ptrdiff_t tindex = 3*cindex;
            acceleration->get_cdata()[tindex+0][0] +=
                    this->kk->kx[xindex]*this->v[1]->get_cdata()[tindex+0][1];
            acceleration->get_cdata()[tindex+0][1] +=
                   -this->kk->kx[xindex]*this->v[1]->get_cdata()[tindex+0][0];
            acceleration->get_cdata()[tindex+1][0] +=
                    this->kk->ky[yindex]*this->v[1]->get_cdata()[tindex+1][1];
            acceleration->get_cdata()[tindex+1][1] +=
                   -this->kk->ky[yindex]*this->v[1]->get_cdata()[tindex+1][0];
            acceleration->get_cdata()[tindex+2][0] +=
                    this->kk->kz[zindex]*this->v[1]->get_cdata()[tindex+2][1];
            acceleration->get_cdata()[tindex+2][1] +=
                   -this->kk->kz[zindex]*this->v[1]->get_cdata()[tindex+2][0];
        }
    }
    );
    /* 12 23 31 */
    this->v[1]->real_space_representation = true;
    this->cvelocity->RLOOP (
                [&](ptrdiff_t rindex,
                    ptrdiff_t xindex,
                    ptrdiff_t yindex,
                    ptrdiff_t zindex){
        //ptrdiff_t tindex = 3*rindex;
        for (int cc=0; cc<3; cc++)
            this->v[1]->rval(rindex,cc) = \
                this->cvelocity->rval(rindex,cc)*this->cvelocity->rval(rindex,(cc+1)%3) / this->cvelocity->npoints;
    }
    );
    this->v[1]->dft();
    this->kk->template dealias<rnumber, THREE>(this->v[1]->get_cdata());
    this->kk->CLOOP(
                [&](const ptrdiff_t cindex,
                    const ptrdiff_t xindex,
                    const ptrdiff_t yindex,
                    const ptrdiff_t zindex,
                    const double k2){
        if (k2 <= this->kk->kM2)
        {
            ptrdiff_t tindex = 3*cindex;
            acceleration->get_cdata()[tindex+0][0] +=
                    (this->kk->ky[yindex]*this->v[1]->get_cdata()[tindex+0][1] +
                     this->kk->kz[zindex]*this->v[1]->get_cdata()[tindex+2][1]);
            acceleration->get_cdata()[tindex+0][1] +=
                  - (this->kk->ky[yindex]*this->v[1]->get_cdata()[tindex+0][0] +
                     this->kk->kz[zindex]*this->v[1]->get_cdata()[tindex+2][0]);
            acceleration->get_cdata()[tindex+1][0] +=
                    (this->kk->kz[zindex]*this->v[1]->get_cdata()[tindex+1][1] +
                     this->kk->kx[xindex]*this->v[1]->get_cdata()[tindex+0][1]);
            acceleration->get_cdata()[tindex+1][1] +=
                  - (this->kk->kz[zindex]*this->v[1]->get_cdata()[tindex+1][0] +
                     this->kk->kx[xindex]*this->v[1]->get_cdata()[tindex+0][0]);
            acceleration->get_cdata()[tindex+2][0] +=
                    (this->kk->kx[xindex]*this->v[1]->get_cdata()[tindex+2][1] +
                     this->kk->ky[yindex]*this->v[1]->get_cdata()[tindex+1][1]);
            acceleration->get_cdata()[tindex+2][1] +=
                  - (this->kk->kx[xindex]*this->v[1]->get_cdata()[tindex+2][0] +
                     this->kk->ky[yindex]*this->v[1]->get_cdata()[tindex+1][0]);
        }
    }
    );
    if (this->kk->layout->myrank == this->kk->layout->rank[0][0])
        std::fill_n((rnumber*)(acceleration->get_cdata()), 6, 0.0);
    this->kk->template force_divfree<rnumber>(acceleration->get_cdata());
}


/*****************************************************************************/




/*****************************************************************************/
/* finally, force generation of code for single precision                    */
template class vorticity_equation<float, FFTW>;
template class vorticity_equation<double, FFTW>;
/*****************************************************************************/

