import os
import sys
import getpass
import subprocess

import numpy as np
import h5py
import matplotlib.pyplot as plt

import TurTLE
import TurTLE.DNS
from TurTLE._code import _code

username = getpass.getuser()
homefolder = os.path.expanduser('~')
cpp_location = None

import pathlib
cpp_location = pathlib.Path(__file__).parent.resolve()


class SCDNS(TurTLE.DNS):
    def __init__(
            self,
            **kwargs):
        TurTLE.DNS.__init__(self, **kwargs)
        self.list_of_particle_codes = []
        self.extra_parameters = {
                'kraichnan_field' : {},
                'NSVE_Stokes_particles' : {},
                'NSVEp_extra_sampling' : {}}
        self.check_current_vorticity_exists = False
        return None
    def write_src(
            self):
        self.version_message = (
                '/***********************************************************************\n' +
                '* this code automatically generated by TurTLE\n' +
                '* version {0}\n'.format(TurTLE.__version__) +
                '***********************************************************************/\n\n\n')
        self.include_list = [
                '"full_code/main_code.hpp"',
                '<cmath>']
        self.main = """
            int main(int argc, char *argv[])
            {{
                bool fpe = (
                    (getenv("BFPS_FPE_OFF") == nullptr) ||
                    (getenv("BFPS_FPE_OFF") != std::string("TRUE")));
                return main_code< {0} >(argc, argv, fpe);
            }}
            """.format(self.dns_type + '<{0}>'.format(self.C_field_dtype))
        self.includes = '\n'.join(
                ['#include ' + hh
                 for hh in self.include_list])
        self.name = 'solver_comparison'
        self.dns_type = 'solver_comparison'
        self.definitions = ''
        self.definitions += open(
            os.path.join(
                cpp_location, 'abstract_fluid_solver.hpp'), 'r').read()
        self.definitions += open(
            os.path.join(
                cpp_location, 'fluid_solver.hpp'), 'r').read()
        self.definitions += open(
            os.path.join(
                cpp_location, 'solver_comparison.hpp'), 'r').read()
        self.definitions += open(
            os.path.join(
                cpp_location, 'solver_comparison.cpp'), 'r').read()
        with open(self.name + '.cpp', 'w') as outfile:
            outfile.write(self.version_message + '\n\n')
            outfile.write(self.includes + '\n\n')
            outfile.write(self.definitions + '\n\n')
            outfile.write(self.main + '\n')
        return None
    def generate_default_parameters(self):
        TurTLE.DNS.generate_default_parameters(self)
        self.extra_parameters['solver_comparison'] = {
                'max_err_estimate' : 1e-5,
                'max_err_relative_estimate' : 1e-4,
                }
        return None

    def write_par(
            self,
            iter0 = 0):
        assert (self.parameters['niter_todo'] % self.parameters['niter_stat'] == 0)
        assert (self.parameters['niter_todo'] % self.parameters['niter_out']  == 0)
        assert (self.parameters['niter_out']  % self.parameters['niter_stat'] == 0)
        #os.chdir(os.path.join(self.work_dir, self.simname))
        _code.write_par(self, iter0 = iter0)
        with h5py.File(self.get_data_file_name(), 'r+') as ofile:
            ofile['code_info/exec_name'] = self.name
            kspace = self.get_kspace()
            for k in kspace.keys():
                ofile['kspace/' + k] = kspace[k]
            nshells = kspace['nshell'].shape[0]
            kspace = self.get_kspace()
            nshells = kspace['nshell'].shape[0]
            vec_stat_datasets = []
            for ff in ['velocity', 'vorticity']:
                for tt in ['1', '2']:
                    for solver1 in ['NSE', 'NSVE']:
                        for solver2 in ['NSE', 'NSVE']:
                            vec_stat_datasets += [
                        ff + '_' + solver1 + '_vs_' + solver2 + '_err' + tt,
                        ff + '_' + solver1 + '_vs_' + solver2 + '_err' + tt + '_relative']
            scal_stat_datasets = []
            for k in vec_stat_datasets:
                time_chunk = 2**20//(8*3*3*nshells)
                time_chunk = max(time_chunk, 1)
                ofile.create_dataset('statistics/spectra/' + k + '_' + k,
                                     (1, nshells, 3, 3),
                                     chunks = (time_chunk, nshells, 3, 3),
                                     maxshape = (None, nshells, 3, 3),
                                     dtype = np.float64)
                time_chunk = 2**20//(8*4*10)
                time_chunk = max(time_chunk, 1)
                a = ofile.create_dataset('statistics/moments/' + k,
                                     (1, 10, 4),
                                     chunks = (time_chunk, 10, 4),
                                     maxshape = (None, 10, 4),
                                     dtype = np.float64)
                time_chunk = 2**20//(8*4*self.parameters['histogram_bins'])
                time_chunk = max(time_chunk, 1)
                ofile.create_dataset('statistics/histograms/' + k,
                                     (1,
                                      self.parameters['histogram_bins'],
                                      4),
                                     chunks = (time_chunk,
                                               self.parameters['histogram_bins'],
                                               4),
                                     maxshape = (None,
                                                 self.parameters['histogram_bins'],
                                                 4),
                                     dtype = np.int64)
            ofile['checkpoint'] = int(0)
        for member_id in [1, 2, 4]:
            for fluid_type in ['NSVE', 'NSE']:
                simname = self.simname + fluid_type + '_' + str(member_id)
                cc = TurTLE.DNS()
                cc.prepare_launch(
                        args = [
                            fluid_type,
                            '--simname', simname,
                            '--wd', self.work_dir])
                for kk in cc.parameters.keys():
                    cc.parameters[kk] = self.parameters[kk]
                # member-specific parameters
                cc.parameters['dt'] = self.parameters['dt'] / member_id
                cc.parameters['niter_todo'] = self.parameters['niter_todo'] * member_id
                cc.parameters['niter_out'] = self.parameters['niter_out'] * member_id
                cc.parameters['niter_stat'] = self.parameters['niter_stat'] * member_id
                cc.write_par()
        return None
    def prepare_launch(
            self,
            **kwargs):
        opt = TurTLE.DNS.prepare_launch(self, **kwargs)
        return opt
    def launch(
            self,
            args = [],
            **kwargs):
        # if no --simname argument is added to args with the same value as self.ensname (or self.simname),
        # self.simname is overwritten with the DNS default inside of prepare_launch.
        # to avoid multiple assignments of self.ensname, we enforce the correct naming scheme here.
        assert('--simname' not in args)
        args += ['--simname', self.simname]

        opt = self.prepare_launch(args = args)
        if not os.path.exists(self.get_data_file_name()):
            self.write_par()
        self.run(
                nb_processes = opt.nb_processes,
                nb_threads_per_process = opt.nb_threads_per_process,
                njobs = opt.njobs,
                hours = opt.minutes // 60,
                minutes = opt.minutes % 60,
                no_submit = opt.no_submit,
                no_debug = not opt.debug)
        return None
    def run_sanity_check(
            self,
            fname = None):
        if type(fname) == type(None):
            fname = self.simname
        ofile = open(fname + '.tex', 'w')
        ofile.write('\\documentclass{article}\n')
        ofile.write('\\usepackage{graphicx}\n')
        ofile.write('\\usepackage[a4paper,'
                  + 'bindingoffset = 0cm,'
                  + 'left   = 1cm,'
                  + 'right  = 1cm,'
                  + 'top    = 1cm,'
                  + 'bottom = 2cm]{geometry}\n')
        ofile.write('\\begin{document}\n')
        members = self.get_member_DNS_list()
        for cc in members:
            cc.read_parameters()
            print(cc.simname)
            print(cc.parameters)
            cc.compute_statistics()
            cc.plot_basic_stats(fig_dir = './figs')
            ofile.write('\\verb|' + cc.work_dir + '|\n\n')
            ofile.write('\\verb|' + cc.simname + '|\n\n')
            ofile.write('forcing type is \\verb|' + cc.parameters['forcing_type'] + '|\n\n')
            ofile.write('\\includegraphics[width=\\textwidth]{'
                      + './figs/stats_' + cc.simname + '.pdf}\n')
            ofile.write('\\newpage\n')
        ofile.write('\\end{document}\n')
        ofile.close()
        subprocess.check_call(['pdflatex', fname + '.tex'])
        return members
    def get_member_DNS_list(self):
        """
            Method called "get_member_DNS_list" to make it obvious that
            this returns a list of DNS objects.
        """
        member_list = [ TurTLE.DNS(
                        simname = '{0}NSVE_{1}'.format(self.simname, i),
                        work_dir = self.work_dir)
                        for i in [1, 2, 4]]
        member_list += [ TurTLE.DNS(
                        simname = '{0}NSE_{1}'.format(self.simname, i),
                        work_dir = self.work_dir)
                        for i in [1, 2, 4]]
        return member_list
    def read_error_statistics(
            self):
        df = self.get_data_file()
        for ff in ['velocity', 'vorticity']:
            for ee_type in ['NSVE_vs_NSVE',
                            'NSVE_vs_NSE',
                            'NSE_vs_NSVE',
                            'NSE_vs_NSE']:
                for ii in [1, 2]:
                    bname = ff + '_' + ee_type + '_err{0}'.format(ii)
                    self.statistics['error/moments/{0}'.format(bname)] = \
                            df['statistics/moments/{0}'.format(bname)][()]
                    self.statistics['error/moments/{0}_relative'.format(bname)] = \
                            df['statistics/moments/{0}_relative'.format(bname)][()]
                    self.statistics['error/histograms/{0}'.format(bname)] = \
                            df['statistics/histograms/{0}'.format(bname)][()]
                    self.statistics['error/histograms/{0}_relative'.format(bname)] = \
                            df['statistics/histograms/{0}_relative'.format(bname)][()]
        df.close()
        return None
    def plot_error_convergence(
            self,
            fname = None,
            err_norm = 1.0):
        if type(fname) == type(None):
            fname = self.simname + '_err_convergence'
        dt = np.array([self.parameters['dt'], self.parameters['dt']/2])
        f = plt.figure(figsize = (8, 8))
        a = f.add_subplot(111)
        dashes = {
                'NSE_vs_NSE' :  (None, None),
                'NSE_vs_NSVE':  (1, 1),
                'NSVE_vs_NSE':  (2, 3),
                'NSVE_vs_NSVE': (3, 4),
                }
        for suffix in ['NSE_vs_NSE',
                       'NSE_vs_NSVE',
                       'NSVE_vs_NSE',
                       'NSVE_vs_NSVE']:
            for ff in ['velocity', 'vorticity']:
                dname = ff + '_' + suffix
                ee = np.array(
                    [self.statistics['error/moments/{0}_err1'.format(dname)][-1, 2, 3]**0.5,
                     self.statistics['error/moments/{0}_err2'.format(dname)][-1, 2, 3]**0.5])
                a.plot(dt, ee, label = dname, dashes = dashes[suffix])
                fit_information = np.polyfit(
                        np.log(dt), np.log(ee), 1)
                convergence_order = fit_information[0]
                assert(convergence_order > 2.75)
        a.plot(dt, dt**3, color = 'black', dashes = (2, 2),
                label = '$\\tau^3$')
        a.set_xscale('log')
        a.set_yscale('log')
        a.set_xlabel('timestep $\\tau$')
        a.legend(loc = 'best')
        f.tight_layout()
        f.savefig(fname + '.pdf')
        plt.close(f)
        return None
    def plot_error_history(
            self,
            fname = None,
            err_norm = 1.0):
        if type(fname) == type(None):
            fname = self.simname + '_err_history'
        dt = np.array([self.parameters['dt'], self.parameters['dt']/2])
        f = plt.figure(figsize = (8, 8))
        a = f.add_subplot(111)
        dashes = {
                'NSE_vs_NSE' :  (None, None),
                'NSE_vs_NSVE':  (1, 1),
                'NSVE_vs_NSE':  (2, 3),
                'NSVE_vs_NSVE': (3, 4),
                }
        for suffix in ['NSE_vs_NSE',
                       'NSE_vs_NSVE',
                       'NSVE_vs_NSE',
                       'NSVE_vs_NSVE']:
            for ff in ['velocity', 'vorticity']:
                dname = ff + '_' + suffix
                ee = self.statistics['error/moments/{0}_err1'.format(dname)][:, 2, 3]**0.5
                a.plot(ee, label = dname, dashes = dashes[suffix])
        a.set_yscale('log')
        a.legend(loc = 'best')
        f.tight_layout()
        f.savefig(fname + '.pdf')
        plt.close(f)
        return None

def main():
    bla = SCDNS()
    niterations = 64
    bla.launch(
            ['solver_comparison',
             '--np', '1',
             '--ntpp', '1',
             '--niter_todo', '{0}'.format(niterations),
             '--niter_out', '{0}'.format(niterations),
             '--niter_stat', '8',
             '--precision', 'double',
             '--forcing_type', 'linear',
             #'--famplitude', '0',
             '--dt', '0.1',
             '--field_random_seed', '2']
            + sys.argv[1:])
    bla.read_parameters()
    #mm = bla.run_sanity_check()
    bla.read_error_statistics()
    bla.plot_error_convergence(err_norm = 1) #mm[0].statistics['Uint'])
    #bla.plot_error_history(err_norm = 1) #mm[0].statistics['Uint'])
    return None

if __name__ == '__main__':
    main()
