import os
import sys
import argparse
import getpass
from packaging import version

username = getpass.getuser()
homefolder = os.path.expanduser('~')

here = os.path.normcase(__file__)

cpp_location = os.path.dirname(here)

import numpy as np
import h5py

import TurTLE
#assert(version.parse(TurTLE.__version__) > version.parse('3.16'))
import TurTLE._base
import TurTLE.DNS
from TurTLE._code import _code


class GCDNS(TurTLE.DNS):
    def __init__(
            self,
            **kwargs):
        self.dns_type = 'many_NSVEcylinder_collisions'
        TurTLE.DNS.__init__(self, **kwargs)
        self.extra_parameters = {} #Added this here to get it to work
        self.extra_parameters[self.dns_type] = {
            'niter_part' : int(10),
            'niter_part_fine_period' : int(10),
            'niter_part_fine_duration' : int(0),
            'nparticles' : np.array([100],dtype=int),
            'tracers0_integration_steps' : int(2),
            'tracers0_neighbours' : int(3),
            'tracers0_smoothness' : int(2),
            'tracers0_enable_p2p' : int(1),
            'tracers0_enable_inner' : int(0),
            'tracers0_enable_vorticity_omega' : int(0),
            'tracers0_cutoff' : np.array([0.5],dtype='float64'),
            'tracers0_lambda' : np.array([100.0],dtype='float64'),
            'tracers0_beta0' : np.array([0.0],dtype='float64'),
            'tracers0_beta1' : np.array([0.0],dtype='float64'),
            'tracers0_D_rho' : np.array([1.0],dtype='float64'),
            'tracers0_u_sed' : np.array([1.0],dtype='float64'),
            'l_factor' : float(1.0),
            't_factor' : float(1.0),
            'many_integration_steps' : np.array([2]).astype(int),
                }
        return None
    def write_src(
            self):
        self.version_message = (
                '/***********************************************************************\n' +
                '* this code automatically generated by TurTLE\n' +
                '* version {0}\n'.format(TurTLE.__version__) +
                '***********************************************************************/\n\n\n')
        self.include_list = [
                '"base.hpp"',
                '"scope_timer.hpp"',
                '"fftw_interface.hpp"',
                '"full_code/main_code.hpp"',
                '"full_code/NSVEparticles.hpp"',
                '<cmath>',
                '<iostream>',
                '<hdf5.h>',
                '<string>',
                '<cstring>',
                '<fftw3-mpi.h>',
                '<omp.h>',
                '<cfenv>',
                '<cstdlib>']
        self.main = """
            int main(int argc, char *argv[])
            {{
                bool fpe = (
                    (getenv("BFPS_FPE_OFF") == nullptr) ||
                    (getenv("BFPS_FPE_OFF") != std::string("TRUE")));
                return main_code< {0} >(argc, argv, fpe);
            }}
            """.format(self.dns_type + '<{0}>'.format(self.C_field_dtype))
        self.includes = '\n'.join(
                ['#include ' + hh
                 for hh in self.include_list])
        self.definitions += open(
                os.path.join(
                    cpp_location, 'particles_inner_computer_sedimenting_rods.hpp'), 'r').read()
        self.definitions += open(
                os.path.join(
                    cpp_location, 'particles_inner_computer_sedimenting_rods.cpp'), 'r').read()
        self.definitions += open(
            os.path.join(
                cpp_location, 'many_NSVEcylinder_collisions.hpp'), 'r').read()
        self.definitions += open(
            os.path.join(
                cpp_location, 'many_NSVEcylinder_collisions.cpp'), 'r').read()
        with open(self.name + '.cpp', 'w') as outfile:
            outfile.write(self.version_message + '\n\n')
            outfile.write(self.includes + '\n\n')
            outfile.write(self.definitions + '\n\n')
            outfile.write(self.main + '\n')
        self.check_current_vorticity_exists = True
        return None

    def generate_tracer_state(
            self,
            rseed = None,
            species = 0,
            integration_steps = None,
            ncomponents = 3
            ):
        try:
            if type(integration_steps) == type(None):
                integration_steps = self.NSVEp_extra_parameters['tracers0_integration_steps']
            if 'tracers{0}_integration_steps'.format(species) in self.parameters.keys():
                integration_steps = self.parameters['tracers{0}_integration_steps'.format(species)]
            if self.dns_type == 'static_field_with_cylinder_collisions' or self.dns_type == 'NSVEcylinder_collisions' or self.dns_type == 'many_NSVEcylinder_collisions':
                ncomponents = 6
            with h5py.File(self.get_checkpoint_0_fname(), 'a') as data_file:
                nn = self.parameters['nparticles'][species]
                if not 'tracers{0}'.format(species) in data_file.keys():
                    data_file.create_group('tracers{0}'.format(species))
                    data_file.create_group('tracers{0}/rhs'.format(species))
                    data_file.create_group('tracers{0}/state'.format(species))
                data_file['tracers{0}/rhs'.format(species)].create_dataset(
                        '0',
                        shape = (integration_steps, nn, ncomponents,),
                        dtype = np.float64)
                dset = data_file['tracers{0}/state'.format(species)].create_dataset(
                        '0',
                        shape = (nn, ncomponents,),
                        dtype = np.float64)
                if not type(rseed) == type(None):
                    np.random.seed(rseed)
                cc = int(0)
                batch_size = int(1e6)
                def get_random_phases(npoints):
                    return np.random.random(
                                    (npoints, 3))*2*np.pi
                def get_random_versors(npoints):
                    if self.parallel_rods:
                        bla = np.zeros((npoints, 3),dtype='float64')
                        bla[:,1]=1.0
                    else:
                        bla = np.random.normal(
                                size = (npoints, 3))
                        bla  /= np.sum(bla**2, axis = 1)[:, None]**.5
                    return bla
                while nn > 0:
                    if nn > batch_size:
                        dset[cc*batch_size:(cc+1)*batch_size, :3] = get_random_phases(batch_size)
                        if dset.shape[1] == 6:
                            dset[cc*batch_size:(cc+1)*batch_size, 3:] = get_random_versors(batch_size)
                        nn -= batch_size
                    else:
                        dset[cc*batch_size:cc*batch_size+nn, :3] = get_random_phases(nn)
                        if dset.shape[1] == 6:
                            dset[cc*batch_size:cc*batch_size+nn, 3:] = get_random_versors(nn)
                        nn = 0
                    cc += 1
        except Exception as e:
            print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!WARNING')
            print('WARNING: when generating particles, caught the exception ', e)
            print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!WARNING')
        return None

    def launch(
            self,
            args = [],
            test_collisions = False,
            parallel_rods = False,
            **kwargs
            ):
        self.parallel_rods = parallel_rods
        opt = self.prepare_launch(args = args)
        if not os.path.exists(self.get_data_file_name()):
            self.write_par()
            self.generate_initial_condition(opt = opt)
            for species in range(self.parameters['many_integration_steps'].shape[0]):
                self.generate_tracer_state(
                        species = species,
                        ncomponents = 6,
                        integration_steps = self.parameters['many_integration_steps'][species-1]
                        )
                with h5py.File(self.get_particle_file_name(), 'a') as particle_file:
                    particle_file.create_group('tracers{0}/position'.format(species))
                    particle_file.create_group('tracers{0}/orientation'.format(species))
                    particle_file.create_group('tracers{0}/velocity'.format(species))
                    particle_file.create_group('tracers{0}/velocity_gradient'.format(species))
                    particle_file.create_group('tracers{0}/acceleration'.format(species))
                with h5py.File(os.path.join(self.work_dir, self.simname + '_colliding_particle_subset.h5'), 'a') as subset_file:
                    subset_file.create_group('subset_tracers{0}/labels'.format(species))
                    subset_file.create_group('subset_tracers{0}/position'.format(species))
                    subset_file.create_group('subset_tracers{0}/orientation'.format(species))
                    subset_file.create_group('subset_tracers{0}/velocity'.format(species))
                    subset_file.create_group('subset_tracers{0}/velocity_gradient'.format(species))
                    subset_file.create_group('subset_tracers{0}/acceleration'.format(species))
        # Create collisions file
        if not os.path.exists(os.path.join(self.work_dir, self.simname + '_collisions.h5')):
            with h5py.File(os.path.join(self.work_dir, self.simname + '_collisions.h5'), 'w') as ofile:
                for i in range(self.parameters['many_integration_steps'].size):
                    ofile.create_group('/collisions{0}'.format(i))
                    ofile.create_group('/new_collisions{0}'.format(i))
                    ofile.create_group('/same_collisions{0}'.format(i))
                    ofile.create_group('/old_collisions{0}'.format(i))
        if not os.path.exists(os.path.join(self.work_dir, self.simname + '_collisions_PP.h5')):
            with h5py.File(os.path.join(self.work_dir, self.simname + '_collisions_PP.h5'), 'w') as ofile:
                for i in range(self.parameters['many_integration_steps'].size):
                    ofile.create_group('/new_collisions{0}'.format(i))
        self.run(
                nb_processes = opt.nb_processes,
                nb_threads_per_process = opt.nb_threads_per_process,
                njobs = opt.njobs,
                hours = opt.minutes // 60,
                minutes = opt.minutes % 60,
                no_submit = opt.no_submit,
                no_debug = opt.no_debug)
        return None

def main():
    bla = GCDNS()
    bla.launch(['many_NSVEcylinder_collisions'] + sys.argv[1:])
    return None

if __name__ == '__main__':
    main()
